/*

DESCRIPTION:    Main control firmware for TS350 inverter:
    + Full bridge inverter
    + Thermal protection & fan management
    + Main input power relay & inrush protection
    + Under & over-volt protection
    + Sine section on/off control
    + Over-current protection
    + Power monitoring & overload protection
    + Service port / console interface

SYNTAX:         n/a

KNOWN BUGS:     n/a

COPYRIGHT:      (c) 2019, Brian Cornell, all rights reserved.

DEPENDENCIES:
    main.h - configuration constants for main.c
    usart.c - RS232 service module
    usart.h - RS232 service module
    config.h - PIC config register values

HW REVISION:    1

HW RELEASE:     190314

VERSION:        2.0

MODIFICATION HISTORY:

    19 Mar 19, BWC, created release.
    

*/


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pic.h>
#include <xc.h>
#include <binary.h>
#include "config.h"
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Make sure _XTAL_FREQ is defined at project level.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifndef _XTAL_FREQ
#include "_xtal_freq.h"
#endif
#include "usart.h"
#include "main.h"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                              Module Data
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
invcf Config;                               //Configuration data
bit cmdGood;                                //TRUE if command successful
bit statMonitor;                            //TRUE if console monitoring on
bit prnResult;                              //TRUE if console access detected
bit fanDisable;                             //TRUE if fan has been disabled at console
bit noSample;                               //TRUE until first sample set ready after boot
unsigned char monTimer;                     //Used by monitor feature & maintained by timer8 ISR

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Non-volatile parameter storage & management.  nvData resides in program
 * flash and ramData in ram.  Initialization of both must match!  These are
 * system defaults and enable recovery after NVM corruption or to restore 
 * from console.  Refer to detailed explanation of NVM management at beginning 
 * of function Initialize.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const unsigned char nvData[FLASHROWSZ] = {S_DF_FLASH_0, S_DF_FLASH_1, \
S_DF_FLASH_2, S_DF_FLASH_3, S_DF_FLASH_4, S_DF_FLASH_5, S_DF_FLASH_6, \
S_DF_FLASH_7, S_DF_FLASH_8, S_DF_FLASH_9, S_DF_FLASH_10, S_DF_FLASH_11, \
S_DF_FLASH_12, S_DF_FLASH_13, S_DF_FLASH_14, S_DF_FLASH_15, S_DF_FLASH_16, \
S_DF_FLASH_17, S_DF_FLASH_18, S_DF_FLASH_19, S_DF_FLASH_20, S_DF_FLASH_21, \
S_DF_FLASH_22, S_DF_FLASH_23, S_DF_FLASH_24, S_DF_FLASH_25, S_DF_FLASH_26, \
S_DF_FLASH_27, S_DF_FLASH_28, S_DF_FLASH_29};
unsigned char ramData[FLASHROWSZ] = {S_DF_FLASH_0, S_DF_FLASH_1, \
S_DF_FLASH_2, S_DF_FLASH_3, S_DF_FLASH_4, S_DF_FLASH_5, S_DF_FLASH_6, \
S_DF_FLASH_7, S_DF_FLASH_8, S_DF_FLASH_9, S_DF_FLASH_10, S_DF_FLASH_11, \
S_DF_FLASH_12, S_DF_FLASH_13, S_DF_FLASH_14, S_DF_FLASH_15, S_DF_FLASH_16, \
S_DF_FLASH_17, S_DF_FLASH_18, S_DF_FLASH_19, S_DF_FLASH_20, S_DF_FLASH_21, \
S_DF_FLASH_22, S_DF_FLASH_23, S_DF_FLASH_24, S_DF_FLASH_25, S_DF_FLASH_26, \
S_DF_FLASH_27, S_DF_FLASH_28, S_DF_FLASH_29};



/*Function main.  Control program for TS350 inverter.  Manages boot, operation, 
 * fault handling, user interface, and shutdown.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void main(void)
{
    char *CmdBuffer;                    //Points to console service command buffer
    unsigned char cmdCount,             //Console service command parsing
            avgCount = S_IN_SAMPLES,    //Counts number of samples taken
            ovpCount = 0,               //Power hysteresis counter
            sinCount = 0,               //Sine hysteresis counter
            ledmap;                     //LED code rendering
    unsigned int icsAvgSample = 0,      //Sum of inverter current samples
            invAvgSample = 0,           //Sum of inverter voltage samples
            ovcCount = 0,               //Current hysteresis counter
            pwr_b,                      //Boost power, watts
            pwr_100,                    //100% power, watts
            pwr_75,                     //75%  power, watts
            pwr_50,                     //50% power, watts
            pwr_25;                     //25%  power, watts
    volts icsSample,                    //Per sample inverter current data
            invSample;                  //Per sample inverter voltage data


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *                           INITIALIZATION
     * 
     * Boot MCU, initialize peripherals, and get ready for operation.  
     * Afterwards take a pause for environment to settle.
     * 
     * Bit object noSample is set to indicate that no active samples are 
     * populated in their respective objects.  Until it is set, the run-time 
     * loop will not execute beyond the data collection blocks.  This is by 
     * design and ensures that the remainder of the loop is working with valid
     * data.  Once the first valid samples are ready noSample is cleared and
     * loop execution proceeds normally.  This is a one-time event at boot.
     * 
     * Ensure that function objects icsAvgSample & invAvgSample are initialized 
     * to zero to ensure they represent accurate sums for first sample set.
     * 
     * The 'pwr' variables are assigned the power values, in watts, representing
     * the power levels to drive the front panel meter.  They are a hack for
     * using constants which the compiler doesn't handle properly since they
     * overflow a signed 16-bit integer (even though they are evaluated against
     * unsigned 16-bit).
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Initialize();
    __delay_ms(S_TM_INT);
    noSample = TRUE;
    pwr_100 = S_LM_PWR_100;
    pwr_b = ((pwr_100 * 11) / 10);
    pwr_75 = ((pwr_100 * 75) / 100);
    pwr_50 = ((pwr_100 * 50) / 100);
    pwr_25 = ((pwr_100 * 25) / 100);

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *             Run-Time Loop Prioritization & Watchdog Management
     * 
     * Once entered, the run-time loop's only exit points are a reboot or power
     * cycle.  The loop consists of several discrete blocks responsible for 
     * data collection, monitoring, and control.  They are organized in order
     * of criticality and/or dependency of subsequent blocks.  Categorically the
     * order is:
     *      + Data collection
     *      + Thermal management
     *      + Run-time monitoring & control
     *      + Boot (autostart)
     *      + NVM status recording
     *      + Status LED display
     *      + Fault recovery (optional)
     *      + Console service
     * 
     * The watchdog timer is expected to be on in the runtime loop regardless of
     * state.  It should be set to a value that accounts for the longest loop
     * time with at least a 100% margin.
     * 
     * If it must be disabled in anticipation of a long delay then it is the
     * responsibility of that code block to reenable.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    SWDTEN = TRUE;

    while (TRUE)
    {
        CLRWDT();                           //Clear watchdog timer

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                          DATA COLLECTION
         * 
         * Scan operating environment.
         * 
         * Various system voltages are read first.  Those that require averaging
         * are stored in temporary objects for summing & subsequent averaging.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReadVoltage(PIN_I_IVDCPV, S_RT_INV, &invSample);        //Inverter bridge voltage
        ReadVoltage(PIN_I_IVCSSH, S_RT_ICS, &icsSample);        //Current sense S&H
        ReadVoltage(PIN_I_INVTMP, S_RT_TMP, &Config.tmp);       //Inverter temp

        getCOGtime();                                           //Fetch duty cycle


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Calculate average of measured samples for:
         *  + Inverter voltage (objects invAvgSample, invSample)
         *  + Inverter current (objects icsAvgSample, icsSample)
         * 
         * On each loop sum the measured value and, when sample count is reached
         * average the value.
         * 
         * Over a number of samples the inverter voltage is expected to be 
         * stable so no further processing is done on it.
         * 
         * However, the inverter current will have a significant AC component
         * resembling the AC output.  Ideally an RMS value would be calculated 
         * but the processing overhead is significant and evaluation of the 
         * approach documented here provides a reasonable analog.
         * 
         * Inverter current is averaged in the same loop as voltage.  To obtain
         * the analog of RMS current the averaged value is reduced by 
         * Config.icsr percent (reduction factor).  This is then scaled for the 
         * current sense circuit gain and the integer quotient assigned to 
         * Config.amps as the inverter current (straight integer; no 'vmv' 
         * format).
         * 
         * This approach works because the averaged current sample represents
         * the voltage across the sample & hold capacitor which itself is
         * integrating the pulse-pulse samples from the current sense circuit.
         * However, the sense circuit delivers an exaggerated slope with a large
         * spike (parasitics) at the end of each cycle which inflates the 
         * measured s&h voltage.  The reduction factor compensates for this.
         * Testing in the first prototype shows an accuracy of about 0/-1.5A 
         * which for the intended purposes is sufficient.
         * 
         * The power is relatively a straight-forward DC volts * amps 
         * calculation.  Using the volts integer only value would depredate 
         * accuracy but using the full vmv format risks overflow of the 16-bit
         * integer.  So the voltage is reduced to integer and one decimal place
         * in the 'vmv' format.  To obtain this the vmv value is divided by 100,
         * for example 12350 (e.g. 12.35 volts) would be represented as 123.
         * This hybrid value is then multiplied by amps and divided again by 10
         * (to eliminate the 10th's place from the volts value) to obtain an 
         * integer only value for power.
         * 
         * Note:  when the inverter is not running (specifically the COG is off),
         * the current S&H circuit voltage floats and will indicate large 
         * current values.  If not cleared this will interfere with the run-time
         * checks immediately after start before a new sample set is taken.  To
         * prevent this the ics sample is cleared when the status is not RUN.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        invAvgSample += invSample.adc;                          //Inverter bridge voltage
        icsAvgSample += icsSample.adc;                          //Current sense S&H
        if (++avgCount > S_LM_SAMPLES)
        {
            //Hack to clear noise readings from S&H circuit when COG is off
            if (Config.sts != STS_RUN) icsAvgSample = 0;

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Calculate averages
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            CalcAvg(invAvgSample, S_LM_SAMPLES, S_RT_INV, &Config.inv);
            CalcAvg(icsAvgSample, S_LM_SAMPLES, S_RT_ICS, &Config.ics);

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Calculate inverter amperes & power
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            Config.amps = ((Config.ics.vmv - ((Config.ics.vmv * Config.icsr) \
                    / S_RT_PERCENT)) / S_RT_ICS_MV);
            Config.pwr = ((Config.amps * (Config.inv.vmv / S_RT_P_VOLTS)) \
                    / S_RT_POWER);

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Assign value to equivalent power.  This value is the actual 
             * inverter power when operating above the setpoint voltage and,
             * when below is a calculated equivalent power that derates the
             * inverter to protect it from excessive current.  The equivalent
             * power value is calculated to avoid additional conditional logic
             * in the overload and power meter sections.
             * 
             * The setpoint voltage refers to an input voltage that is reasonably
             * expected to be available in the real world - e.g. automotive 13~
             * 13.8V and the inverter can supply rated power (350W output) to 
             * the sine section with some operating margin for boost (thermally
             * limited).
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (Config.inv.vmv < S_VL_INV_SET)
                Config.pwr_e = (((((S_VL_INV_SET - Config.inv.vmv) \
                        / S_RT_P_VOLTS) * Config.amps) / S_RT_POWER) + \
                        Config.pwr);
            else Config.pwr_e = Config.pwr;

            avgCount = S_IN_SAMPLES;                            //Reset counter
            icsAvgSample = 0;                                   //Clear sample
            invAvgSample = 0;                                   //Clear sample
            noSample = FALSE;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Need valid sample @ boot before continuing
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (noSample) continue;


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                        THERMAL MANAGEMENT
         * 
         * The fan is cycled according to sensed temperature regardless of run 
         * state provided it hasn't been manually disabled.
         * 
         * The NTC's voltage read by the ADC is compared to a known value
         * representing the thresholds.  The voltage-temperature relationship
         * is expected to be inverse (e.g. high temp=low volts).
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (!(fanDisable))
        {
            if ((Config.tmp.vmv < Config.tfon) && (!(PIN_O_FAN))) \
                PIN_O_FAN = TRUE;
            else if ((Config.tmp.vmv > Config.tfof) && (PIN_O_FAN)) \
                PIN_O_FAN = FALSE;
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                    RUN-TIME MONITORING & CONTROL
         * 
         * Consists of three sections.  The first checks parameters regardless
         * of run state and sets Config.sts accordingly (first one set sticks).
         * The second are grouped together under status RUN conditional since 
         * only applicable when the unit is in RUN state.  Some employ a simple 
         * P-D hysteresis loop (current, power) to accommodate transients or 
         * starting currents.
         * 
         * The RUN section employes a waterfall loop so the first error sets 
         * status & breaks loop.  Hence, the most important / time critical 
         * checks should be done first.
         * 
         * In both monitoring sections no action beyond setting status is taken,
         * however, a status set by the first section will inhibit operation.
         * 
         * The last section evaluates any status set by the previous two and
         * if the unit is operating the inverter (and by design sine) and relay
         * are turned off.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.sts < STS_CODE_ERR)
        {
            //High voltage
            if (Config.inv.vmv > S_VL_INV_MAX) Config.sts = STS_BUS_OVLO;
            //Low voltage
            else if (Config.inv.vmv < S_VL_INV_MIN) Config.sts = STS_BUS_UVLO;
            //High temperature
            else if (Config.tmp.vmv < Config.tovh) Config.sts = STS_HIGH_TEMP;
        }

        if (Config.sts == STS_RUN)
        {
            while (TRUE)
            {
                //Over-current.  Simple P-D loop.
                if (Config.cogc < Config.oldc) ++ovcCount; else if (ovcCount) \
                    --ovcCount;
                if (ovcCount > S_LM_OVC_HYS)
                {
                    ovcCount = 0;
                    Config.sts = STS_OVC_OCLO;
                    break;
                }

                //Overload (power)
                if (Config.pwr_e > S_LM_PWR_MAX) ++ovpCount; else if (ovpCount) \
                    --ovpCount;
                if (ovpCount > S_LM_OVP_HYS)
                {
                    ovpCount = 0;
                    Config.sts = STS_OVERLOAD;
                    break;
                }

                //Sine.
                if ((PIN_O_SINE) && !(PIN_I_SINE)) ++sinCount; 
                else if (sinCount) --sinCount;
                if (sinCount > S_LM_SIN_HYS)
                {
                    sinCount = 0;
                    Config.sts = STS_SIN_CMD;
                    break;
                }

                break;
            }
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * SAFETY SHUT-DOWN:  shut inverter down if running and error code is
         * set by checks.  Stopping the inverter, by function, also stops 
         * sine.
         * 
         * For safety the relay is also turned off to eliminate power
         * to the inverter bridge.
         * 
         * This action, unless _FAULT_RECOVERY options are set, effectively 
         * disables the unit until power cycled or console intervention.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((invStatus()) && (Config.sts != STS_RUN))
        {
            InverterOFF();
            setRelay(FALSE);
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                        BOOT (AUTOSTART)
         * 
         * This block is responsible for completing the boot process in the run-
         * time loop and establishing the proper run-time state based on NVM 
         * data.
         * 
         * The boot process does not proceed until valid voltage samples are
         * available (refer to initialization block at top of main function).
         * 
         * Watchdog is disabled due to the long delays between inverter & sine 
         * start.  If autostart is set the DC bus relay is switched on and then 
         * the inverter is started.  A delay is added to allow the sine section 
         * to boot and quiesce.  A failure in any of these steps will abort 
         * autostart; it's expected that the function where start failed has set 
         * Config.sts to the appropriate error code.  If the sine section fails
         * to start the inverter is also stopped for safety.
         * 
         * If autostart is disabled this code block still executes to complete
         * the boot process and set the STOP status code.
         * 
         * Per the runtime loop design, the watchdog is always reenabled at 
         * completion.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.sts == STS_BOOT)
        {
            SWDTEN = FALSE;                     //Make sure disabled

            if (ramData[FLASH_AR_AUTO] != S_BL_AUTOD)
            {
                setRelay(TRUE);
                __delay_ms(S_TM_RELAY);
                if (InverterON())
                {
                    if (ramData[FLASH_AR_AUTO] == S_BL_AUTOS)
                    {
                        __delay_ms(S_TM_START);
                        if (setSine(TRUE)) Config.sts = STS_RUN;
                        else InverterOFF();
                    }
                    else Config.sts = STS_RUN;
                }
            }
            else Config.sts = STS_STOP;

            SWDTEN = TRUE;                      //Enable watchdog timer
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                       NVM STATUS RECORDING
         * Record last error status to NVM for service.  Don't record under-volt
         * lockout (UVLO) since this will happen on normal shutdown.  Before 
         * committing, ensure that VCC is greater than minimum to prevent NVM 
         * corruption if power is lost.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((Config.sts >= STS_CODE_ERR) && (ramData[FLASH_AR_STS] != Config.sts))
        {
            ReadVoltage(PIN_I_IVVCCV, S_RT_VCC, &Config.vcc);
            if ((Config.sts != STS_BUS_UVLO) && (Config.vcc.vmv > S_LM_VCC_NVM))
            {
                ramData[FLASH_AR_STS] = Config.sts;
                writeFlashRow((uint16_t) nvData, ramData);
            }
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                         STATUS LED DISPLAY
         * 
         * Set the front panel LEDs according to the status & power level.  To 
         * display a status code each respective pin's bit is masked and shifted
         * to the bit 0 position to set its pin.
         * 
         * Refer to the 'STATUS CODES & LED MAP' in the include file for details
         * on the status code masks.
         * 
         * Object ledmap is assigned a copy of the LED map for the status code
         * since the StsList array resides in program flash with slower access
         * times.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (Config.sts == STS_RUN)
        {
            //Set run LED
            PIN_O_FAULT = FALSE;
            PIN_O_RUN = TRUE;

            //Set 100% LED to 100% or boost
            PIN_O_BOOST = (Config.pwr_e > pwr_b);
            PIN_O_L100 = (PIN_O_BOOST ^ (Config.pwr_e > pwr_100));

            //Remaining power LEDs function as bar graph
            PIN_O_L75 = (Config.pwr_e > pwr_75);
            PIN_O_L50 = (Config.pwr_e > pwr_50);
            PIN_O_L25 = (Config.pwr_e > pwr_25);
        }
        else
        {
            ledmap = StsList[Config.sts].ledmap;        //Copy to RAM

            //Set pins according to status map to display fault & code
            PIN_O_FAULT = (ledmap & S_MK_FAULT);
            PIN_O_RUN = ((ledmap & S_MK_RUN) >> 1);
            PIN_O_BOOST = ((ledmap & S_MK_BOOST) >> 2);
            PIN_O_L100 = ((ledmap & S_MK_L100) >> 3);
            PIN_O_L75 = ((ledmap & S_MK_L75) >> 4);
            PIN_O_L50 = ((ledmap & S_MK_L50) >> 5);
            PIN_O_L25 = ((ledmap & S_MK_L25) >> 6);
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                           FAULT RECOVERY
         * 
         * This is where code should be placed if there are conditions that the 
         * unit should attempt automated recovery from.
         * 
         * Example:  low input voltage is detected, unit shuts down, then power
         * source recovers.  Unit could resume operation and avoid / reduce
         * operator intervention.
         * 
         * Since it's expected that this design will be used in unstable 
         * environments it's preferable to not attempt recovery from this or 
         * other conditions.  The designer can enable this by uncommenting
         * the '_FAULT_RECOVERY' constant in conditional compilation section
         * of the include file.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifdef _FAULT_RECOVERY
        if (Config.sts > STS_CODE_OK)
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Recovery from inverter DC (+) bus low voltage.  Works by changing
             * status to BOOT which allows boot section of run loop to recover
             * system to RUN status defined by NVM.
             *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if ((Config.sts == STS_BUS_UVLO) && (Config.inv.vmv > S_VL_INV_RSTART))
                Config.sts = STS_BOOT;
        }
#endif


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                     CONSOLE PARAMETER MONITORING
         * 
         * Works like Unix 'top' utility to display system data on console at ~ 
         * 1S intervals:
         *  + Inverter DC+
         *  + Vcc sense
         *  + Bridge temp (NTC volts)
         *  + Duty cycle
         *  + S&H current sense
         * 
         * Bit-flag statMonitor controls when monitoring is on and is set with 
         * console command.  For code space efficiency a common format string
         * is used with all parameters except the last one which includes the
         * line feed (instead of tab) to avoid the overhead of a separate call.
         * The single-byte monTimer, maintained by the system timing ISR, times
         * the update interval which is set by constant S_TM_MONINTVL.
         * 
         * Care should be taken with the use of this feature due to the overhead 
         * it imposes.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (statMonitor)
        {
            if (!(monTimer))
            {
                printf(fmtMonvd, lblInv, Config.inv.volt, Config.inv.mv);
                printf(fmtMonvd, lblTemp, Config.tmp.volt, Config.tmp.mv);
                printf(fmtMonvd, lblDc, (unsigned int) (Config.cogt >> 4),\
                        (unsigned int) (Config.cogt & 0x0F));
                printf(fmtMonvd, lblAmp, Config.amps, 0);
                printf(fmtMonlt, lblPwr, Config.pwr, 0);
                monTimer = S_TM_MONINTVL;           //Reset timer
            }
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         *                         CONSOLE SERVICE
         * 
         * Process user input.  Status bit prnResult is set the first time 
         * console input is detected.  When clear the command result macros 
         * printCmdResult & printCmdSys don't send data to the service port to 
         * reduce noise, resource consumption, etc.  Once console access is 
         * detected prnResult remains TRUE until power is cycled.
         * 
         * IMPORTANT:  do not place any code expected to execute on each loop
         * iteration beyond this point.  The console service handler should 
         * always be placed at bottom of run loop.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (usartStatRx() != USARTSTSRXRDY) continue;
        CmdBuffer = usartGetRx();
        prnResult = TRUE;                       //Detected console access

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * The 'GET' GPIB style commands are evaluated first for response
         * priority.
         *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        //Get input amps
        if (!(stricmp(CmdBuffer, CmdList[CMDGETAMP].str)))
            printf(fmtGetv, Config.amps);
        //Get inverter duty cycle
        else if (!(stricmp(CmdBuffer, CmdList[CMDGETDYC].str)))
            printf(fmtGetvd, (unsigned int) (Config.cogt >> 4),\
                    (unsigned int) (Config.cogt & 0x0F));
        //Get inverter temp
        else if (!(stricmp(CmdBuffer, CmdList[CMDGETEMP].str)))
            printf(fmtGetvd, Config.tmp.volt, Config.tmp.mv);
        //Get inverter input volts
        else if (!(stricmp(CmdBuffer, CmdList[CMDGETVOL].str)))
            printf(fmtGetvd, Config.inv.volt, Config.inv.mv);
        //Get input watts
        else if (!(stricmp(CmdBuffer, CmdList[CMDGETWAT].str)))
            printf(fmtGetv, Config.pwr);
        //Print help info
        else if (!(strcmp(CmdBuffer, CmdList[CMDDSPHLP].str)))
        {
            printf(FMT_HELP_INTRO, msgName, msgMod, msgHwr, VERMAJOR, VERMINOR, \
                    VERBUILD, msgCW);
            for (cmdCount = 0; (cmdCount < NUMOFCMDS); ++cmdCount)\
                printf("    %s=%s\n", CmdList[cmdCount].str, CmdList[cmdCount].desc);
        }
        //Display system status
        else if (!(stricmp(CmdBuffer, CmdList[CMDDSPSTS].str))) DisplayStatus();
        //Toggle monitor on / off
        else if (!(stricmp(CmdBuffer, CmdList[CMDTOGMON].str)))
        {
            clearCmdResult();
            if (statMonitor) usartSetEcho(TRUE);
            else
            {
                monTimer = S_TM_MONINTVL;
                usartSetEcho(FALSE);
            }
            statMonitor ^= TRUE;
            printTogResult(statMonitor);
        }
        //Set duty cycle
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETDCY].str, CMDSIZSTR)))\
            setDutyCycle(CmdBuffer);
        //Set ampere calculation reduction factor
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETRAT].str, CMDSIZSTR)))\
            setAmpsReduction(CmdBuffer);
        //Set blanking time
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETBLK].str, CMDSIZSTR)))\
            setBlanking(CmdBuffer);
        //Configure autostart
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETAUT].str, CMDSIZSTR)))\
            setAutostart(CmdBuffer);
        //Toggle inverter on / off
        else if (!(stricmp(CmdBuffer, CmdList[CMDTOGINV].str)))
        {
            if (invStatus()) InverterOFF();
            else InverterON();
        }
        //Toggle sine on / off
        else if (!(stricmp(CmdBuffer, CmdList[CMDTOGSIN].str)))
        {
            if (sineStatus()) setSine(FALSE);
            else setSine(TRUE);
        }
        //Toggle relay on / off
        else if (!(stricmp(CmdBuffer, CmdList[CMDTOGRLY].str)))
        {
            if (relayStatus()) setRelay(FALSE);
            else setRelay(TRUE);
        }
        //Set per-cycle current limit
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETCUR].str, CMDSIZSTR)))\
            setCurrentLimit(CmdBuffer);
        //Restore NVM defaults
        else if (!(stricmp(CmdBuffer, CmdList[CMDRSTDFL].str)))
        {
            clearCmdResult();
            writeFlashRow((uint16_t) nvData, NULL);
            printCmdResult();
        }
        //Reset MCU
        else if (!(stricmp(CmdBuffer, CmdList[CMDRSTPIC].str)))
        {
            //Reset system; first do clean stop if running
            if (!(COG1ASD0bits.ASE)) InverterOFF();
            SWDTEN = FALSE;             //Disable so watchdog doesn't erroneously reboot
            PCON = S_LM_PCON_BT;        //Clear PCON bits to clear error trap on boot
            __delay_ms(S_TM_RST);
            RESET();
        }
        //Clear NVM error
        else if (!(stricmp(CmdBuffer, CmdList[CMDCLRERR].str)))
        {
            clearCmdResult();
            ramData[FLASH_AR_STS] = STS_UND;
            writeFlashRow((uint16_t) nvData, ramData);
            printCmdResult();
        }
        //Toggle fan disable
        else if (!(stricmp(CmdBuffer, CmdList[CMDTOGFAN].str)))
        {
            clearCmdResult();
            fanDisable ^= TRUE;
            printTogResult(fanDisable ^ TRUE);
        }
        //Set console character echo on / off.  Anything other than 0 turns on.
        else if (!(strnicmp(CmdBuffer, CmdList[CMDSETECH].str, CMDSIZSTR)))
        {
            rmvSpaces(CmdBuffer);   //Clear white space
            if (*CmdBuffer == CMDECHOOF) usartSetEcho(FALSE);
            else usartSetEcho(TRUE);
        }
        //Set fan OFF temp
        else if (!(strnicmp(CmdBuffer, CmdList[CMDTMPFOF].str, CMDSIZSTR)))\
            setTemps(CmdBuffer, FLASH_AR_FAN_F_L);
        //Set fan ON temp
        else if (!(strnicmp(CmdBuffer, CmdList[CMDTMPFON].str, CMDSIZSTR)))\
            setTemps(CmdBuffer, FLASH_AR_FAN_O_L);
        //Set OVERHEAT temp
        else if (!(strnicmp(CmdBuffer, CmdList[CMDTMPOVH].str, CMDSIZSTR)))\
            setTemps(CmdBuffer, FLASH_AR_OVH_L);
        //Gibberish entered
        else printf(msgHuh);

        usartSetRx(ENABLE);             //Look for new user input
    }
} //End function main


/* Function DisplayStatus.  Prints status & configuration:
 * 
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void DisplayStatus()
{
    ReadVoltage(ADC_CH_DAC_4, S_RT_VDD, &Config.vdd);
    ReadVoltage(PIN_I_IVVCCV, S_RT_VCC, &Config.vcc);

    printf("\n%s%s, NVM=%s, %s=%s", lblSys, StsList[Config.sts].desc,\
            StsList[ramData[FLASH_AR_STS]].desc, CmdList[CMDTOGMON].str, \
            ((statMonitor) ? msgOn : msgOff));
    printf("\nMCU: STKOVF=%u, STKUNF=%u, RWDT=%u, RMCLR=%u, RI=%u, POR=%u, BOR=%u, TO=%u, PD=%u",\
            (unsigned int) STKOVF, (unsigned int) STKUNF, (unsigned int) nRWDT,\
            (unsigned int) nRMCLR, (unsigned int) nRI, (unsigned int) nPOR,\
            (unsigned int) nBOR, (unsigned int) nTO, (unsigned int) nPD);
    printf("\nControl/Ack: %s=%c, %s=%s, %s=%s, %s=%s/%s, %s=", \
            CmdList[CMDTOGFAN].str, ((fanDisable) ? CMDDSABLE : CMDENABLE), \
            CmdList[CMDTOGRLY].str, ((relayStatus()) ? msgOn : msgOff), \
            CmdList[CMDTOGINV].str, ((invStatus()) ? msgOn : msgOff), \
            CmdList[CMDTOGSIN].str, ((sineStatus()) ? msgOn : msgOff), \
            ((sineAck()) ? msgOn : msgOff), CmdList[CMDSETAUT].str);
    if (ramData[FLASH_AR_AUTO] == S_BL_AUTOI) printf(msgInv);
    else if (ramData[FLASH_AR_AUTO] == S_BL_AUTOS) printf(msgSine);
    else printf(msgOff);
    printf("\nTiming:  %s=%u%%, measured=%u.%uuS, %s=%u", \
            CmdList[CMDSETDCY].str, Config.odcp, (unsigned int) (Config.cogt >> 4), \
            (unsigned int) (Config.cogt & 0x0F), CmdList[CMDSETBLK].str, \
            Config.btim);
    printf("\nThermal:  %s%u, fan=%s, %s=%u, %s=%u, %s=%u", lblTemp, \
            Config.tmp.vmv, ((PIN_O_FAN) ? msgOn : msgOff), \
            CmdList[CMDTMPFON].str, Config.tfon, CmdList[CMDTMPFOF].str,
            Config.tfof, CmdList[CMDTMPOVH].str, Config.tovh);
    printf("\nCurrent: %s=%d, s&h=%u.%3.3u, %s=%d, %s%d, %s%d", \
            CmdList[CMDSETCUR].str, (int) ((DAC1REFH << 8) | (DAC1REFL)), \
            Config.ics.volt, Config.ics.mv, CmdList[CMDSETRAT].str, \
            (int) Config.icsr, lblAmp, Config.amps, lblPwr, Config.pwr);
    printf("\nVoltages: %s%u.%3.3u, Vcc=%u.%3.3u, Vdd=%u.%3.3u",\
            lblInv, Config.inv.volt, Config.inv.mv, Config.vcc.volt, Config.vcc.mv,\
            Config.vdd.volt, Config.vdd.mv);

    printf(msgNewl);

    return;
} //End function DisplayStatus


/* Function setSine.  Sets output pin to enable / disable digital sine bridge
 * (AC output).  Once the pin is set an acknowledgment is expected; failure
 * to receive results in an error return and the command pin is cleared.
 * 
 * Returns:  boolean TRUE if ack received, FALSE if not.  If FALSE then output
 *           pin is set false regardless of cmd and Config.sts contains error
 *           code.
 *
 * Parameters:  cmd, non-0=on, 0=off
 *
*/
bit setSine(unsigned char cmd)
{
    clearCmdResult();

    if (cmd)                                //Enable
    {
        PIN_O_SINE = TRUE;
        //Wait S_LM_SIN_ON x S_TM_ISN_ON increments for ack.
        for (cmd = S_LM_SIN_ON; (!(PIN_I_SINE) && (cmd)); --cmd, __delay_us(S_TM_SIN_ON));
    }
    else                                    //Disable
    {
        PIN_O_SINE = FALSE;
        //Wait S_LM_SIN_OFF x S_TM_SIN_OFF increments for ack.
        for (cmd = S_LM_SIN_OFF; ((PIN_I_SINE) && (cmd)); --cmd, __delay_us(S_TM_SIN_OFF));
    }

    //If cmd==0 assume timeout occurred; set output off & error return.
    if (!(cmd))
    {
        PIN_O_SINE = FALSE;
        Config.sts = STS_SIN_CMD;
        setCmdErr();
    }

    printTogResult(PIN_O_SINE);

    return(cmdGood);
} //End function setSine


/* Function setRelay.  Controls the PWM output driving the floating flyback gate 
 * drive for SSR on/off control.  The positive (+) pole SSR requires a floating
 * gate drive that is provided by a simple flyback circuit operating at a fixed
 * duty cycle.
 * 
 * PWM3 is a 10-bit PWM with the time base provided by timer2 and is expected
 * to run at 100kHz.
 * 
 * The PWM is configured to be always on and on/off control is achieved by 
 * changing the duty cycle from zero (off) to CFG_PWM_3_DCH_ON.  The two LSBs of
 * duty cycle aren't used; so only the DCH register is needed.  This simplifies
 * turn on/off.  Since the gate drive is a high impedance load (ignoring start)
 * and the duty cycle is fixed at a low value it's not critical the Microchip
 * recommended start procedure.  This makes turn-on/off as simple as loading
 * the proper value into the DCH register.
 * 
 * There is no diagnostic to confirm that the relay is operating once the PWM
 * is on.
 * 
 * The relay will not turn off if the inverter is running.  This is necessary
 * to prevent potential damage to the inverter start / auxiliary supply circuit
 * and likely opening of its fuse.
 * 
 * Returns:     none
 *
 * Parameters:  cmd=non-0, relay on
 *              cmd=0, relay off
 *
*/
void setRelay(unsigned char cmd)
{
    clearCmdResult();

    if (cmd) PWM3DCH = CFG_PWM_3_DCH_ON;            //Enable
    else
    {
        if (Config.sts == STS_RUN) setCmdErr();
        else PWM3DCH = CFG_PWM_3_DCH_OFF;           //Disable
    }

    printTogResult(PWM3DCH);

    return;
} //End function setRelay


/* Function setAutostart.  Sets autostart and updates NV memory.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setAutostart(char *setstr)
{
    unsigned char start;

    clearCmdResult();

    start = ramData[FLASH_AR_AUTO];                     //Get current setting
    rmvSpaces(setstr);                                  //Get thru white space
    if (*setstr < 97) *setstr += CMDLOWCAS;             //Change to lower case for eval

    if (*setstr == CMDINVERT) start = S_BL_AUTOI;       //Inverter only
    else if (*setstr == CMDINSINE) start = S_BL_AUTOS;  //Inverter & sine
    else if (*setstr == CMDDSABLE) start = S_BL_AUTOD;  //Disable        
    else setCmdErr();

    //Update NVM if changed
    if (ramData[FLASH_AR_AUTO] != start)
    {
        ramData[FLASH_AR_AUTO] = start;
        writeFlashRow((uint16_t) nvData, ramData);
    }

    printCmdResult();

    return;
} //End function setAutostart


/* Function setDutyCycle.  Stores & sets the duty cycle.  If running the change
 * is done gradually for smooth operation.
 * 
 * Returns:  none.  Config.odcp & Config.oldc contain the new duty cycle.
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setDutyCycle(char *setstr)
{
    unsigned int dcp;                       //User entered duty cycle as %
    unsigned int oldcSave;                  //Value of oldc prior to change

    clearCmdResult();

    //Process white space & perform bounds check
    rmvSpaces(setstr);
    dcp = atoi(setstr);
    if ((dcp < S_DC_MIN_P) || (dcp > S_DC_MAX_P)) setCmdErr();
    else if (dcp != Config.odcp)
    {
        oldcSave = Config.oldc;
        Config.odcp = dcp;
        Config.oldc = (((dcp * CFG_PWM_5_PERIOD) / S_RT_PERCENT) + 1);

        //If operating move to new DC now
        if (!(COG1ASD0bits.ASE)) chgDutyCycle(oldcSave, Config.oldc);

        //Update NVM
        ramData[FLASH_AR_OLDC_P] = (unsigned char) Config.odcp;
        ramData[FLASH_AR_OLDC] = (unsigned char) Config.oldc;
        writeFlashRow((uint16_t) nvData, ramData);
    }

    printCmdResult();

    return;
} //End function setDutyCycle


/* Function setAmpsReduction.  Processes a console entered integer value 
 * representing the percent to reduce the averaged ICS current calculation and
 * stores the updated value in NVM.  The input value is bounds checked.
 * 
 * Returns:  none.  Config.icsr contains new reduction value.
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setAmpsReduction(char *setstr)
{
    unsigned char p_reduce;

    clearCmdResult();

    //Process white space & perform bounds check
    rmvSpaces(setstr);
    p_reduce = (unsigned char) atoi(setstr);
    if ((p_reduce < S_LM_ICS_MIN) || (p_reduce > S_LM_ICS_MAX)) setCmdErr();
    else if (p_reduce != Config.icsr)
    {
        Config.icsr = p_reduce;
        ramData[FLASH_AR_ICS] = Config.icsr;
        writeFlashRow((uint16_t) nvData, ramData);
    }

    printCmdResult();

    return;
} //End function setAmpsReduction


/* Function setTemps.  Validates & stores system temperatures:
 *      + Fan on
 *      + Fan off
 *      + Overheat (shutdown)
 * 
 * The user entered voltage is expected to be in vmv format (refer to 
 * explanation '16-bit 'VMV' values' in main.h).  Values are stored in object
 * Config in the vmv format.
 * 
 * Returns:  none
 *
 * Parameters:  setstr=NULL terminated string entered at console
 *              set_t=NVM address for low byte representing temperature to
 *              update (e.g. constant FLASH_AR_XXX_L).  EXTREME CARE must be
 *              exercised to ensure that a valid address is passed since no 
 *              bounds checking is performed.
 *
*/
void setTemps(char *setstr, char set_t)
{
    unsigned int utemp;

    clearCmdResult();

    //Process white space & perform bounds check
    rmvSpaces(setstr);
    utemp = atoi(setstr);
    if ((utemp < S_LM_TMP_H) || (utemp > S_LM_TMP_L)) setCmdErr();
    else
    {
        if (set_t == FLASH_AR_FAN_O_L) Config.tfon = utemp;
        else if (set_t == FLASH_AR_FAN_F_L) Config.tfof = utemp;
        else if (set_t == FLASH_AR_OVH_L) Config.tovh = utemp;

        //Update NVM only if data changed
        if (utemp != (unsigned int) ((((unsigned int) ramData[set_t + 1]) \
                << S_SZ_BYTE) | ((unsigned int) ramData[set_t])))
        {
            ramData[set_t] = (unsigned char) (utemp & S_MK_U16_L);
            ramData[set_t + 1] = (unsigned char) ((utemp & S_MK_U16_H) >> S_SZ_BYTE);
            writeFlashRow((uint16_t) nvData, ramData);
        }
    }

    printCmdResult();

    return;
} //End function setTemps


/* Function setBlanking.  Sets the blanking time on rising edge of COG to
 * prevent spurious current sense falling events.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setBlanking(char *setstr)
{
    unsigned int dc;                       //User entered duty cycle

    clearCmdResult();

    //Process white space & perform bounds check
    rmvSpaces(setstr);
    if ((dc = atoi(setstr)) > CFG_PWM_11_PERIOD) setCmdErr();
    else if (dc != Config.btim)
    {
        Config.btim = dc;
        ramData[FLASH_AR_BTIM] = dc;
        PWM11DC = dc;
        PWM11LDCONbits.LDA = TRUE;

        writeFlashRow((uint16_t) nvData, ramData);
    }

    printCmdResult();

    return;
} //End function setBlanking


/* Function setCurrentLimit.  Sets the current limit by updating the DAC with 
 * user-supplied value.  The entered value is bounds checked.  When changed
 * the updated value is stored in NVM.
 * 
 * Returns:  none
 *
 * Parameters:  char *setstr points to NULL terminated string entered at console.
 *
*/
void setCurrentLimit(char *setstr)
{
    unsigned int limit;                     //New current limit value

    clearCmdResult();

    //Process white space & perform bounds check
    rmvSpaces(setstr);
    limit = atoi(setstr);
    if ((limit < S_LM_CLLOW) || (limit > S_LM_CLHIG)) setCmdErr();
    else if (limit != Config.clim)
    {
        Config.clim = limit;
        ramData[FLASH_AR_CLIM_H] = (unsigned char) (limit >> S_SZ_BYTE);
        ramData[FLASH_AR_CLIM_L] = (unsigned char) (limit & S_MK_U16_L);
        DAC1REFH = ramData[FLASH_AR_CLIM_H];
        DAC1REFL = ramData[FLASH_AR_CLIM_L];
        DAC1LD = TRUE;
        while (DAC1LD);

        writeFlashRow((uint16_t) nvData, ramData);
    }

    printCmdResult();

    return;
} //End function setCurrentLimit


/* Function InverterON.  Starts inverter; if already started it has no effect.
 * No pre-start diagnostic voltage checks are done since this is expected to be
 * handled by the run-time loop and reflected in the value of object Config.sts.
 * The diagnostic exception is a check to make sure the relay is on, since if
 * not the auxiliary / start circuit will fail.
 * 
 * The bootstrap caps are first charged and this also provides an opportunity to
 * check for high-side failures in the bridge (e.g. a shorted high-side MOSFET).
 * 
 * Likewise, the routine will stop the inverter after the bridge is started if a
 * sustained [HW] current limit status exists.
 * 
 * Returns:  TRUE if started.
 *           FALSE if not; Config.sts contains error code except if the relay
 *           was not set on.  In this case the status code remains unchanged.
 *
 * Parameters:  none
 *
*/
bit InverterON()
{
    unsigned char result;

    setCmdErr();

    for (result = 0; (relayStatus()); )
    {
        if (COG1ASD0bits.ASE && ((Config.sts == STS_STOP) || (Config.sts == STS_BOOT)))
        {
            if (result = chargeBScaps()) break;

            PWM5DC = S_DC_MIN;                      //Soft start with min DC
            PWM5LDCONbits.LDA = TRUE;
            COG1ASD0bits.ASE = FALSE;               //Output on
            chgDutyCycle(S_DC_MIN, Config.oldc);    //Increase to set DC

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Delay allows duty cycle to catch-up since during start the 
             * secondary impedance will be very low (bulk caps, inrush of 
             * attached equipment).  Because of this the COG DC will be limited 
             * by the current sense loop.  If DC is not as expected after delay 
             * then something serious is likely wrong and inverter is stopped.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            __delay_ms(S_TM_INVON_COG);
            getCOGtime();
            if (Config.cogc < Config.oldc) result = STS_COG_TIME;

            if (!(result))                          //Start successful
            {
                result = STS_RUN;
                setCmdOK();
            }
            else COG1ASD0bits.ASE = TRUE;           //Start failed, inverter off
        }
        break;
    }

    if (result) Config.sts = result;
    printCmdSys();

    return(cmdGood);
} //End function InverterON


/* Function InverterOFF.  Stops inverter.  The sine command pin is cleared
 * after the inverter is stopped but no confirmation check is performed since
 * its power source was turned off.
 * 
 * Returns:  none.  Config.sts is set to STOP provided it doesn't contain an 
 *           error code.
 *
 * Parameters:  none
 *
*/
void InverterOFF()
{
    COG1ASD0bits.ASE = TRUE;            //Inverter off
    if (Config.sts < STS_CODE_ERR) Config.sts = STS_STOP;   //Don't clear existing error

    //Then disable sine bridge
    PIN_O_SINE = FALSE;

    printCmdSys();

    return;
} //End function InverterOFF


/* Function ReadVoltage.  Reads the value from the analog pin channel, interprets 
 * as a scaled voltage, and stores in the volts object pointed to by rv.
 * 
 * The ADC must be pre-configured to use the FVR x 4 voltage reference (4.096V).
 * 
 * No error checking is performed and only a single sample is taken.  The voltage
 * is stored in 'composite' form for easy integer evaluation of the complete
 * value (e.g. V/mV).  Refer to the explanation '16-bit 'VMV' values' found
 * in the include file.  The discrete volts and mV are also stored in separate 
 * integers.
 * 
 * Returns:  none.  The following values are populated in object rv:
 *      vmv=composite value
 *      volt=integer voltage component
 *      mv=millivolts component
 *      adc=unmodified value read from ADC
 *
 * Parameters:  channel=analog channel.  The pin must be properly configured.
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts.
 *
*/
void ReadVoltage(unsigned char channel, unsigned char scale, volts *rv)
{
    static union {
        unsigned char result[S_SZ_ADC_INT];
        unsigned int value;
    } adc;

    ADCON0 = (unsigned) (channel | S_EN_ADC);   //Connect to channel & enable ADC
    __delay_us(S_TM_ADC_ACQ);                   //Refresh sample & hold cap
    ADCON0bits.GO = TRUE;                       //Start conversion
    while (ADCON0bits.GO);                      //Wait for completion
    adc.result[POS_ADC_HB] = ADRESH;
    adc.result[POS_ADC_LB] = ADRESL;

    rv->adc = adc.value;                        //Save raw ADC value to user object
    adc.value *= scale;                         //Scale value
    rv->vmv = (((adc.value / S_QY_ADC_1V) * S_RT_VOLTS) + \
            ((adc.value % S_QY_ADC_1V) * S_RT_ADC_MV));
    rv->volt = rv->vmv / S_RT_VOLTS;
    rv->mv = rv->vmv % S_RT_VOLTS;

    return;
} //End function ReadVoltage


/* Function CalcAvg.  Calculates the arithmetic mean of a sample set of 
 * previously acquired values.  No bounds checking is performed.
 * 
 * The value to be averaged is expected to represent the sum of raw ADC
 * conversions; they should not be scaled.  The average is calculated and then
 * scaled by this function.  
 * 
 * The voltage is stored in 'composite' form for easy integer evaluation of the 
 * complete value (e.g. V/mV).  Refer to the explanation '16-bit 'VMV' values' 
 * found in the include file.  The discrete volts and mV are also stored in 
 * separate integers.
 * 
 * Returns:  none.  The following values are populated in object rv.  All 
 * represent averaged values:
 *      vmv=composite value
 *      volt=integer voltage component
 *      mv=millivolts component
 *      adc=value read from ADC
 *
 * Parameters:  sum=raw summed ADC values
 *              count=number of summed values
 *              scale=integer representing the scale of the value being read.
 *              rv=pointer to object of type volts.
 *
*/
void CalcAvg(unsigned int sum, unsigned char count, unsigned char scale, \
        volts *rv)
{
    rv->adc = ((sum / count) * scale);
    rv->vmv = (((rv->adc / S_QY_ADC_1V) * S_RT_VOLTS) + \
            ((rv->adc % S_QY_ADC_1V) * S_RT_ADC_MV));
    rv->volt = rv->vmv / S_RT_VOLTS;
    rv->mv = rv->vmv % S_RT_VOLTS;

    return;
} //End function CalcAvg


/* Function getCOGtime.  Calculates the on time of the COG and stores in module
 * object Config (type invcf).  COG on-time capture is performed using timer5 
 * with gate control and single pulse modes.  The base timer configuration is 
 * expected to be completed at boot.
 * 
 * The calculated on-time is stored in a single byte (cogt).  The integer 
 * component of the value is calculated by dividing the timer low byte count by 
 * the count for 1uS.  The modulus of the timer value is then multiplied by 100 
 * and divided by the maximum count (10uS).
 * 
 * Returns:  none.  The following values are populated in Config:
 *      cogc=value from timer, each increment=62.5nS
 *      cogt=upper nibble uS, lower nibble=100's nS
 * 
 * cogc and cogt are set to zero if the timer's lower register exceeds 
 * FGCT_MAXCNT.
 *
 * Parameters:  none
 *
*/
void getCOGtime()
{
    //Clear timer: high byte is not used
    TMR5 = 0;

    T5GGO_nDONE = TRUE;                 //Start capture
    __delay_us(FGCT_WAITTM);            //Fixed wait time for completion
    if (T5GGO_nDONE)                    //Problem or COG is off
    {
        T5GGO_nDONE = FALSE;            //Stop timing
        TMR5L = (FGCT_MAXCNT + 1);      //Set high value & let remaining logic clear
    }

    if (TMR5L > FGCT_MAXCNT)            //Invalid DC - clear values
    {
        Config.cogc = 0;
        Config.cogt = 0;
    }
    else                                //Calculate valid DC
    {
        Config.cogc = TMR5;
        Config.cogt = ((TMR5L / FGCT_1USCNT) << 4);
        Config.cogt |= (((TMR5L % FGCT_1USCNT) * 100) / FGCT_MAXCNT);
    }

    return;
} //End function getCOGtime


/* Function chgDutyCycle.  Moves the duty cycle of the bridge from the start
 * to final value incrementally for smooth operation & transient control.
 * 
 * Returns:     none
 *
 * Parameters:  start=16-bit starting DC value
 *              final=16-bit final DC value
 *
*/
void chgDutyCycle(unsigned int start, unsigned int final)
{
    while (TRUE)
    {
        PWM5DC = start;
        PWM5LDCONbits.LDA = TRUE;
        if (start == final) break;

        if (start < final) ++start; else --start;
        __delay_us(S_TM_DCDELAY);
    }

    return;
} //End function chgDutyCycle


/* Function chargeBScaps.  Charges bootstrap capacitors by alternately turning on
 * the low-side switch of each bridge channel.  The COG is disconnected from
 * low side pin to facilitate and during the process short circuit monitoring
 * is performed.
 * 
 * This function does not perform state checks and expects to be called from a
 * higher level function that ensures the proper operating state.
 * 
 * Short circuit detection is accomplished by directly reading the output of the
 * current sense comparator.  This is done because the over-current event ISR
 * detects a state change only; not a steady state.  Short detection is performed
 * after the charge cycle is expected to complete since the bootstrap capacitors
 * can draw several amperes of current depending on size & current limiting.
 * 
 * Returns:  0 if successful or non-zero value representing error.
 *
 * Parameters:  none
 *
*/
unsigned char chargeBScaps()
{
    unsigned char result;

    for (result = 0; (TRUE); )
    {
        //Charge A-side
        PIN_O_SIDE_A_L = FALSE;
        PPS_P_SIDE_A_L = PPS_M_LAT;             //Take direct control of pin
        PIN_O_SIDE_A_L = TRUE;                  //Turn switch on
        __delay_us(S_TM_BS_CHG);                //BS caps charge
#ifndef _NO_LIMITS
        if (!(S_MP_CSENS)) result = STS_CHL_BSAF;
#endif
        PIN_O_SIDE_A_L = FALSE;                 //Turn switch off
        PPS_P_SIDE_A_L = PPS_M_COG1_D;          //Return pin control to COG
        if (result) break;

        //Charge B-side
        PIN_O_SIDE_B_L = FALSE;
        PPS_P_SIDE_B_L = PPS_M_LAT;             //Take direct control of pin
        PIN_O_SIDE_B_L = TRUE;                  //Turn switch on
        __delay_us(S_TM_BS_CHG);                //BS caps charge
#ifndef _NO_LIMITS
        if (!(S_MP_CSENS)) result = STS_CHL_BSBF;
#endif
        PIN_O_SIDE_B_L = FALSE;                 //Turn switch off
        PPS_P_SIDE_B_L = PPS_M_COG1_C;          //Return pin control to COG

        break;
    }

    return(result);
} //End function chargeBScaps


/* Function writeFlashRow.  Writes a row of flash.  A 16-bit checksum for data
 * is first calculated and stored in the last word of the row.  Hence, the user
 * only has (FLASHROWSZ - 2) bytes available for use.
 * 
 * This function can also be used to restore default NVM settings by leveraging
 * the 'BOOT & NVM PROTECTION AND MANAGEMENT STRATEGY' documented in function
 * Initialize.  To do so this function should be called with the data parameter
 * set to NULL.  When so the checksum word is cleared so that upon the next
 * boot the NVM defaults will be used since NVM data is considered invalid.
 * 
 * Returns:  none
 *
 * Parameters:  addr=14-bit address of row in flash to write.  Must point to
 *                  start of array.
 *              data=pointer to array of data to write to flash -OR-
 *                   NULL to clear the existing checksum but leave other data
 *                   in the row intact.
 *              
 *
*/
void writeFlashRow(uint16_t addr, unsigned char *data)
{
    unsigned int chksum;
    unsigned char count;

    //Writing full row
    if (data)
    {
        count = 0;                              //Start at beginning of row
        chksum = nvmChksum(data);               //Calculate & store checksum
        *(data + FLASH_AR_CSUM_L) = (unsigned char) (chksum & S_MK_U16_L);
        *(data + FLASH_AR_CSUM_H) = \
                (unsigned char) ((chksum & S_MK_U16_H) >> S_SZ_BYTE);

        //Must erase row first for clean write
        eraseFlashRow(addr);
    }
    //Clearing checksum only
    else
    {
        count = FLASH_AR_CSUM_L;                //Start at checksum offset in row
        chksum = 0;                             //Set checksum to zero
        data = (unsigned char *) &chksum;       //Set data address to checksum
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Setup to write data to flash.  Interrupts must be disabled for entire
     * sequence.  Set LWLO to load data to write latches and then to flash on
     * last byte.
     * 
     * Load high address register here & only once.  Update low register for 
     * individual bytes to write within row.
     * 
     * Variable count represents array element to write.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    di();                               //Interrupts disabled for entire sequence.
    WREN = TRUE;                        //Enable write
    CFGS = FALSE;                       //Program memory
    FREE = FALSE;                       //Write data
    LWLO = TRUE;                        //Load data latches
    PMADRH = (unsigned char) (addr >> 8);

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Important:  count must be initialized at beginning of function based on 
     * intent.  It controls number of bytes to write and address offset within
     * NVM row.  data must point to full row or checksum only.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    for (; (count < FLASHROWSZ); ++count, ++data)
    {
        PMADRL = (unsigned char) (addr + count);
        PMDATL = *data;
        unlockSeq();                    //Unlock for write
    }

    LWLO = FALSE;                                   //Write row
    unlockSeq();                                    //Write flash

    //All done:  disable writes & re-enable interrupts
    WREN = FALSE;
    ei();

    return;
} //End function writeFlashRow


/* Function eraseFlashRow.  Erases a row of program flash memory.  Interrupts
 * are disabled during the erase.
 * 
 * Returns:  none
 *
 * Parameters:  addr=14-bit pointer to starting address for row.
 *
*/
void eraseFlashRow(uint16_t addr)
{
    PMADRL = (unsigned char) addr;
    PMADRH = (unsigned char) (addr >> S_SZ_BYTE);

    //Erase row
    di();                               //Disable interrupts for duration of function
    CFGS = FALSE;                       //Program memory
    FREE = TRUE;                        //Erase row
    WREN = TRUE;                        //Enable erase
    unlockSeq();                        //Unlock for erase
    WREN = FALSE;                       //Disable erase
    ei();                               //Enable interrupts

    return;
} //End function eraseFlashRow


/* Function nvmChksum.  Calculates the checksum for a row of data.  The 'row'
 * is an unsigned char array of length FLASHROWSZ since this function is
 * intended to be used data storage in NVM (program flash).
 * 
 * This function is intended to be used in conjunction with writeFlashRow and
 * assumes the checksum resides in the last two bytes of the row.  Hence, when
 * calculating the checksum only (FLASHROWSZ - 2) bytes are examined.
 * 
 * The checksum is a simple sum of byte values in the row starting with offset
 * 0.
 * 
 * Returns:  16-bit checksum value
 *
 * Parameters:  data=pointer to data row
 *              
 *
*/
unsigned int nvmChksum(unsigned char *data)
{
    unsigned int sum;
    unsigned char count;

    for (sum = 0, count=0; (count < (FLASHROWSZ - FLASH_SZ_CSUM)); ++count)
        sum += *(data + count);

    return(sum);
} //End function nvmChksum


/*Function InterruptExec.  MPU interrupt handler.  Manages all enabled
 * interrupts:
 *      + USART Rx (service port RS232 communications)
 *      + Timer8 (real-time background timer)
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void interrupt InterruptExec(void)
{
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Caretaker.  Not sure if this will be needed but leave for future.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (TMR8IE & TMR8IF)
    {
        if (monTimer) --monTimer;
        TMR8IF = FALSE;
    }

    usartISR();

    return;
} //End function InterruptExec


/*Function Initialize.  Configures pins & peripherals on the MPU.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    di();                               //No interrupts during initialization
    Config.sts = STS_BOOT;              //Start with assumption of clean boot...

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * System clock & watchdog config
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OSCCON = CFG_OSC_C;                 //Configure system clock
    while (!(HFIOFL));                  //Wait for oscillator to stabilize
    WDTCON = CFG_WDT_CON;               //Watchdog


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     *           BOOT & NVM PROTECTION AND MANAGEMENT STRATEGY
     *
     * Recovering from CPU/firmware failures (e.g. watchdog, stack OF/UF, etc.)
     * and NVM corruption is critical for safety and field reliability.  While
     * neither should happen unless there's a true failure, statistically it's 
     * possible when operating in environmental extremes or the more likely
     * existence of a firmware bug.
     * 
     * The goal is to keep the unit intact and record the failure state.
     * 
     * DATA MANAGEMENT
     * A single row (32 bytes) of NVM data is used for the storage of tunable
     * parameters (e.g. current limit, thermal management, etc.).  Object nvData
     * resides in flash (NVM) and is initialized with default values.  During
     * normal operation, object ramData contains a copy of nvData and is used
     * to update nvData.  It must also be initialized with the same default
     * values as nvData.
     * 
     * DATA VALIDATION
     * NVM data is validated with a simple 16-bit checksum stored in the last
     * two bytes of nvData.
     * 
     * BOOT PROCESS
     * The CPU boot status is always checked first to ensure that any abnormal
     * condition is identified before being cleared by subsequent execution.  
     * The PCON & STATUS registers are examined for reboot due to watchdog, 
     * stack OF/UF, and possibly other conditions.  If found the status byte
     * (Config.sts) is updated.
     * 
     * The NVM checksum is verified before attempting to load.  The checksum
     * value is read directly from NVM and compared to the NVM's calculated
     * checksum.  If they match nvData is copied to ramData.
     * 
     * If there's a mismatch ramData retains it's default values and the status
     * code will be updated if not already set by the CPU checks.
     * 
     * The remainder of the initialization function continues unaltered and
     * when complete the main function enters the runtime loop.  The inverter
     * will not start if any status code other than BOOT is set.  A code block 
     * in the main function loop is responsible for storing error codes, 
     * resulting from the boot process or otherwise, in NVM.
     * 
     * Since the MCU's architecture requires the entire NVM row to be written
     * it's critical that ramData contains valid data.  This is why it's not
     * updated with nvData when checksum's don't match.  By writing the status
     * code it also ensures that if NVM was truly corrupt it will now contain
     * valid data (provided there's not an MCU flash failure) and there's a 
     * chance of the unit returning to operation when power cycled.
     * 
     * Compile note:  braces are intentionally used even though each side of
     * the evaluation contains only one statement.  During testing it was 
     * found that the compiler did not interpret the evaluation correctly 
     * without them.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (((PCON & S_LM_PCON_EX) != S_LM_PCON_BT) || \
            ((STATUS & S_LM_STAT_EX) != S_LM_STAT_BT)) Config.sts = STS_CPU;

    if (nvmChksum((unsigned char *) nvData) != (unsigned int) ((((unsigned int) \
            nvData[FLASH_AR_CSUM_H]) << S_SZ_BYTE) | ((unsigned int) \
            nvData[FLASH_AR_CSUM_L])))
    {
        if (Config.sts == STS_BOOT) Config.sts = STS_NVM_INV;
    }
    else
    {
        memcpy((void *) ramData, (void *) nvData, FLASHROWSZ);
    }


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Initialize control data
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Config.odcp = (unsigned int) ramData[FLASH_AR_OLDC_P];
    Config.oldc = (unsigned int) ramData[FLASH_AR_OLDC];
    Config.icsr = ramData[FLASH_AR_ICS];
    Config.clim = (unsigned int) ((((unsigned int) ramData[FLASH_AR_CLIM_H])\
            << S_SZ_BYTE) | ((unsigned int) ramData[FLASH_AR_CLIM_L]));
    Config.btim = ramData[FLASH_AR_BTIM];
    Config.tfon = (unsigned int) ((((unsigned int) ramData[FLASH_AR_FAN_O_H])\
            << S_SZ_BYTE) | ((unsigned int) ramData[FLASH_AR_FAN_O_L]));
    Config.tfof = (unsigned int) ((((unsigned int) ramData[FLASH_AR_FAN_F_H])\
            << S_SZ_BYTE) | ((unsigned int) ramData[FLASH_AR_FAN_F_L]));
    Config.tovh = (unsigned int) ((((unsigned int) ramData[FLASH_AR_OVH_H])\
            << S_SZ_BYTE) | ((unsigned int) ramData[FLASH_AR_OVH_L]));
    statMonitor = FALSE;                //Set monitoring off
    prnResult = FALSE;                  //Don't print command results
    fanDisable = FALSE;                 //Fan default is enabled
    OPTION_REGbits.nWPUEN = FALSE;      //Individual pin control over weak pull-ups


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set port configuration.  Set digital latch initial output states.
     * 
     * Set LED status for boot.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PORTA = S_IN_PORTA;
    PORTB = S_IN_PORTB;
    PORTC = S_IN_PORTC;
    //Port A
    TRISA = IO_PORTA;
    ANSELA = AD_PORTA;
    WPUA = WP_PORTA;
    ODCONA = OD_PORTA;
    SLRCONA = SR_PORTA;
    INLVLA = IL_PORTA;
    //Port B
    TRISB = IO_PORTB;
    ANSELB = AD_PORTB;
    WPUB = WP_PORTB;
    ODCONB = OD_PORTB;
    SLRCONB = SR_PORTB;
    INLVLB = IL_PORTB;
    //Port C
    TRISC = IO_PORTC;
    ANSELC = AD_PORTC;
    WPUC = WP_PORTC;
    ODCONC = OD_PORTC;
    SLRCONC = SR_PORTC;
    INLVLC = IL_PORTC;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Set the Boost LED on so there is some indication that the unit has power
     * if it hangs before getting to the run loop.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PIN_O_BOOST = TRUE;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Map peripheral pins
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PPS_P_SIDE_A_H = PPS_M_COG1_A;          //COG1 channel A
    PPS_P_SIDE_A_L = PPS_M_COG1_D;          //COG1 channel D
    PPS_P_SIDE_B_H = PPS_M_COG1_B;          //COG1 channel B
    PPS_P_SIDE_B_L = PPS_M_COG1_C;          //COG1 channel C
    PPS_P_PWM3_SSR = PPS_M_PWM3_SSR;        //PWM3
    T5GPPS = PPS_M_TMR_5G;                  //Timer5 gate


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Initialize USART for command console
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    usartSetRx(ENABLE);


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure timer2 as time base for PWM5
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    T2CLKCON = CFG_TMR_2_CLK;
    T2HLT = CFG_TMR_2_HLT;
    T2RST = CFG_TMR_2_RST;
    T2PR = CFG_TMR_2_PR;
    T2CON = CFG_TMR_2_CON;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure timer8 (caretaker) but don't start
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    T8CLKCON = CFG_TMR_8_CLK;
    T8HLT = CFG_TMR_8_HLT;
    T8RST = CFG_TMR_8_RST;
    T8PR = CFG_TMR_8_PR;
    T8CON = CFG_TMR_8_CON;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure timer5 to measure on-time (duty cycle) of COG1.  Function 
     * getCOGtime is dependent on this.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    T5CON = CFG_TMR_5_C;
    T5GCON = CFG_TMR_5_G;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure FVR & DACs
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    FVRCON = CFG_FVR_C;
    __delay_us(FI_FVR_DLY);

    DAC1CON0 = CFG_DAC_1_C;
    DAC1REFH = ramData[FLASH_AR_CLIM_H];
    DAC1REFL = ramData[FLASH_AR_CLIM_L];
    DAC1LD = TRUE;                      //Load reference value

    DAC4CON0 = CFG_DAC_4_C;
    DAC4REF = CFG_DAC_4_V;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure over-current sense comparator (CM2)
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CM2CON0 = CFG_CM_2_C0;
    CM2CON1 = CFG_CM_2_C1;
    CM2NSEL = CFG_CM_2_N;
    CM2PSEL = CFG_CM_2_P;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure ADC baseline operation.  Clock, +/- references, auto-conversion,
     * and justification don't change.  Enable and default connect to DAC4 to
     * calibrate & obtain Vdd values.
     * 
     * Rule:  no assumptions should be made about which analog channel the ADC
     * is connected to.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    ADCON0 = (ADC_CH_DAC_4 | S_EN_ADC);
    ADCON1 = CFG_ADC_CON_1;
    ADCON2 = CFG_ADC_CON_2;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM3 for SSR on/off control.  PWM is expected to start now
     * but with 0 (zero) DC so SSR is off.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM3DCL = CFG_PWM_3_DCL;
    PWM3DCH = CFG_PWM_3_DCH_OFF;
    PWM3CON = CFG_PWM_3_CON;

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM5 as primary PWM signal
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM5INTE = CFG_PWM_5_INTE;
    PWM5CLKCON = CFG_PWM_5_CLK;
    PWM5LDCON = CFG_PWM_5_LDCON;
    PWM5OFCON = CFG_PWM_5_OFCON;
    PWM5PH = CFG_PWM_5_PHASE;
    PWM5DC = ramData[FLASH_AR_OLDC];
    PWM5PR = CFG_PWM_5_PERIOD;
    PWM5OF = CFG_PWM_5_OFFSET;
    PWM5TMR = CFG_PWM_5_TIMER;
    PWM5CON = CFG_PWM_5_CON;                //Start

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure PWM11 as slave to PWM5 to time current sense blanking
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PWM11INTE = CFG_PWM_11_INTE;
    PWM11CLKCON = CFG_PWM_11_CLK;
    PWM11LDCON = CFG_PWM_11_LDCON;
    PWM11OFCON = CFG_PWM_11_OFCON;
    PWM11PH = CFG_PWM_11_PHASE;
    PWM11DC = ramData[FLASH_AR_BTIM];
    PWM11PR = CFG_PWM_11_PERIOD;
    PWM11OF = CFG_PWM_11_OFFSET;
    PWM11TMR = CFG_PWM_11_TIMER;
    PWM11CON = CFG_PWM_11_CON;              //Start


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure CLC3 as OR gate to control COG1 current sense falling event
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CLC3SEL0 = CFG_CLC_3_SEL_0;
    CLC3SEL1 = CFG_CLC_3_SEL_1;
    CLC3SEL2 = CFG_CLC_3_SEL_2;
    CLC3SEL3 = CFG_CLC_3_SEL_3;
    CLC3GLS0 = CFG_CLC_3_GLS_0;
    CLC3GLS1 = CFG_CLC_3_GLS_1;
    CLC3GLS2 = CFG_CLC_3_GLS_2;
    CLC3GLS3 = CFG_CLC_3_GLS_3;
    CLC3POL = CFG_CLC_3_POL;
    CLC3CON = CFG_CLC_3_CON;
    LC3EN = TRUE;                           //Enable CLC3


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure complementary output generator 1 to drive full bridge
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    COG1DBR = CFG_COG_1_RE_DEADB;
    COG1DBF = CFG_COG_1_FE_DEADB;
    COG1BLKR = CFG_COG_1_RE_BLANK;
    COG1BLKF = CFG_COG_1_FE_BLANK;
    COG1PHR = CFG_COG_1_RE_PHASE;
    COG1PHF = CFG_COG_1_FE_PHASE;
    COG1ASD1 = CFG_COG_1_ASD1;
    COG1ASD0 = CFG_COG_1_ASD0;
    COG1RIS0 = CFG_COG_1_REI0_NONE;
    COG1RIS1 = CFG_COG_1_REI1_PWM5;
    COG1FIS0 = CFG_COG_1_FEI0_NONE;
    COG1FIS1 = CFG_COG_1_FEI1_OLOOP;
    COG1RSIM0 = CFG_COG_1_REM0_NONE;
    COG1RSIM1 = CFG_COG_1_REM1_PWM5;
    COG1FSIM0 = CFG_COG_1_FEM0_NONE;
    COG1FSIM1 = CFG_COG_1_FEM1_OLOOP;
    COG1CON1 = CFG_COG_1_CON1;
    COG1CON0 = CFG_COG_1_CON0;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Setup CLC2; should be done after COG1 config.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    CLC2SEL0 = CFG_CLC_2_SEL_0;
    CLC2SEL1 = CFG_CLC_2_SEL_1;
    CLC2SEL2 = CFG_CLC_2_SEL_2;
    CLC2SEL3 = CFG_CLC_2_SEL_3;
    CLC2GLS0 = CFG_CLC_2_GLS_0;
    CLC2GLS1 = CFG_CLC_2_GLS_1;
    CLC2GLS2 = CFG_CLC_2_GLS_2;
    CLC2GLS3 = CFG_CLC_2_GLS_3;
    CLC2POL = CFG_CLC_2_POL;
    CLC2CON = CFG_CLC_2_CON;
    LC2EN = TRUE;                           //Enable


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Configure current sense S&H amp (OPA2)
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OPA2CON = CFG_OPA_2_C;
    OPA2ORS = CFG_OPA_2_O;
    OPA2NCHS = CFG_OPA_2_N;
    OPA2PCHS = CFG_OPA_2_P;
    OPA2EN = TRUE;                          //Enable


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * Initialization complete:  enable interrupts and start system timer.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PEIE = ENABLE;                      //Enable peripheral interrupts
    ei();                               //Enable interrupts

    //Enable interrupt & start system timer
    TMR8IE = TRUE;
    T8ON = TRUE;

    return;
} //End function Initialize



//End of file