/*

DESCRIPTION:
    Main include file TS350 inverter.

COPYRIGHT:      (c) 2019, Brian Cornell, all rights reserved.

HW REVISION:    1

HW RELEASE:     190314

VERSION:        2.0

MODIFICATION HISTORY:

    4 Apr 19, BWC, moved defintion of _XTAL_FREQ to separate include file
    _xtal_freq.h for proper compile & reference by project source files.

    19 Mar 19, BWC, created release.


*/



/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                      Version constants & definitions
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define VERMAJOR    2
#define VERMINOR    0
#define VERBUILD    50


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                      Necessary includes
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#ifndef UINT_MAX
#include <limits.h>
#endif


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Conditional compilation.  Uncomment directive to enable / disable features.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Uncomment to remove limit checks & not save errors to NVM in run-time loop
//#define _NO_LIMITS

//Uncomment to enable fault recovery options
//#define _FAULT_RECOVERY


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * System wide constants & definitions.  Convention:  S_<TYPE>_<INFO>
 * Types:
 *  + DC=Duty Cycle
 *  + DF=Default value for a configurable parameter
 *  + EN=ENable bit
 *  + RT=RaTio or scaling factor
 *  + LM=LiMiting operational value
 *  + TM=TiMing
 *  + MP=Peripheral port bit MaPping
 *  + CH=character
 *  + IN=INitializing value
 *  + VL=VoLtage
 *  + QY=quantity (count) of something
 *  + CD=command data
 *  + BL=boolean value
 *  + MK=bit MasK
 *  + SZ=size of an object or data type
 * 
 * Info is user defined and should be descriptive.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_TM_INV_F          0x009F          //System inverter PWM frequency
#define S_DC_MIN            0x08            //Minimum allowable duty cycle
#define S_DC_MAX            0X98            //Maximum allowable duty cycle
#define S_DC_RUN            0x98            //Running duty cycle
#define S_DC_MIN_P          ((S_DC_MIN * 100) / (CFG_PWM_5_PERIOD + 1))
#define S_DC_MAX_P          ((S_DC_MAX * 100) / (CFG_PWM_5_PERIOD + 1))
#define S_DC_RUN_P          ((S_DC_RUN * 100) / (CFG_PWM_5_PERIOD + 1))
#define S_EN_ADC            0x01            //Enable bit position in ADCON0 register
#define S_RT_VDD            2               //Vdd ratio rendered by DAC
#define S_RT_INV            11              //Inverter DC+ bus voltage ratio
#define S_RT_VCC            5               //Inverter Vcc voltage ratio
#define S_RT_ICS            1               //Inverter current sense S&H voltage ratio
#define S_RT_TMP            1               //Inverter NTC (temp) sense voltage ratio
#define S_RT_SEN            1               //Ratio for Vcc sense voltage for contro loop
#define S_RT_REF            1               //Ratio for reference voltage for control loop
#define S_VL_INV_MIN        10500           //Minimum inverter DC+ bus voltage
#define S_VL_INV_MAX        16000           //Maximum inverter DC+ bus voltage
#define S_VL_INV_RSTART     12000           //Restart (from UVLO) inverter DC+ bus voltage
#define S_VL_INV_SET        12400           //Set point voltage for calculating equivalent power
#define S_LM_VCC_NVM        9000            //Minimum VCC voltage to allow NVM writes
#define S_TM_RST            200             //# mS to wait before reseting PIC
#define S_TM_INT            300             //# mS for system to stabilize at power-up
#define S_TM_RELAY          25              //# mS for inverter DC+ inrush after relay on
#define S_TM_START          1000            //# mS to wait to start sine after inverter
#define S_MP_CSENS          C2OUT           //CM2 output bit for over-current sense
#define S_LM_CLLOW          1               //Lowest allowable current limit value
#define S_LM_CLHIG          1023            //Highest allowable current limit value
#define S_BL_AUTOD          0               //Autostart disabled
#define S_BL_AUTOI          1               //Autostart inverter only
#define S_BL_AUTOS          2               //Autostart inverter & sine
#define S_LM_OVC_HYS        320             //Hysteresis for over-current event limit
#define S_LM_OVP_HYS        160             //Hysteresis for over-power event limit
#define S_LM_UVL_HYS        160             //Hysteresis for DC bus under volt event limit
#define S_LM_SIN_HYS        10              //Hysteresis for sine acknowledgement fail
#define S_LM_PCON_EX        0xFA            //PCON bits to examine at boot
#define S_LM_PCON_BT        0x18            //PCON expected value at boot
#define S_LM_STAT_EX        0x18            //STATUS bits to examine at boot
#define S_LM_STAT_BT        0x18            //STATUS expected value at boot
#define S_TM_SIN_ON         100             //# uS to wait for sine ON signal
#define S_LM_SIN_ON         20              //# S_TM_SIN_ON periods before sine ON timeout
#define S_TM_SIN_OFF        100             //# uS to wait for sine OFF signal
#define S_LM_SIN_OFF        100             //# S_TM_SIN_OFF periods before sine OFF timeout
#define S_TM_DCDELAY        500             //# uS to wait between each DC increment
#define S_TM_MONINTVL       100             //# SysTimer clock ticks for monitor updates (1S)
#define S_IN_SAMPLES        1               //Starting value for measured samples counter
#define S_LM_SAMPLES        120             //Number of measured samples to average
#define S_TM_INVON_COG      5               //mS delay before measuring COG time on inverter start
#define S_RT_ICS_MV         15              //Current sense mV/A represented as integer
#define S_LM_PWR_MAX        550             //INVERTER maximum allowable power in watts
#define S_LM_PWR_100        450             //INVERTER continuous full power rating in watts
#define S_MK_FAULT          0x01            //Fault indicator on
#define S_MK_RUN            0x02            //Run indicator on
#define S_MK_BOOST          0x04            //Boost indicator on
#define S_MK_L100           0x08            //100% indicator on
#define S_MK_L75            0x10            //75% indicator on
#define S_MK_L50            0x20            //50% indicator on
#define S_MK_L25            0x40            //25% indicator on
#define S_TM_BS_CHG         1               //# uS for bootstrap caps to charge
#define S_SZ_BYTE           8               //Number of bits in a byte
#define S_MK_U16_L          0x00FF          //Mask to access low byte of 16-bit unsigned
#define S_MK_U16_H          0xFF00          //Mask to access high byte of 16-bit unsigned
#define S_LM_TMP_L          4000            //Lowest user-entered temp in vmv
#define S_LM_TMP_H          900             //Highest user-entered temp in vmv
#define S_RT_VOLTS          1000            //Integer component of volts starts in 1000's place of integer
#define S_RT_P_VOLTS        100             //Integer component of volts starts in 100's place of integer for power calcs
#define S_RT_POWER          10              //Integer component of power starts in 10's place
#define S_RT_PERCENT        100             //Used as divisor to calculate % of a dividend value
#define S_RT_ICS_AMPS       25              //Percent (%) to reduce ICS sample average to calculate input amps
#define S_LM_ICS_MIN        1               //Minimum allowed % to reduce ICS sample
#define S_LM_ICS_MAX        50              //Maximum allowed % to reduce ICS sample

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                           16-bit 'VMV' values
 * 
 * Many voltage values are represented as:  (volts * 1000) + (mv * 100) so
 * they can be evaluated as integers with three decimal place precision 
 * with reasonable efficiency on an 8-bit integer platform (ex - 3.426 volts 
 * would be represented as 3426) while still being human readable.  This format 
 * is referred to as 'vmv' (volts-millivolts).
 * 
 * Those that must be stored in NVM at compile time require two bytes and the 
 * 16-bit values must be broken into their respective high/low bytes.  This is 
 * done below and the convention is:
 *      + The 16-bit constant is defined first
 *      + The high & low bytes are defined with the same name and '_H' or '_L'
 *        appended respectively
 *      + The bytes should be assigned to the NVM row in little endian order
 * The compiler will throw an "arithmetic overflow in constant expression" 
 * warning on the assignment lines.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_VL_TMP_OVH        1000            //Inverter NTC voltage representing overheat temp
#define S_VL_TMP_OVH_H      (S_VL_TMP_OVH >> S_SZ_BYTE)
#define S_VL_TMP_OVH_L      (S_VL_TMP_OVH & S_MK_U16_L)
#define S_VL_TMP_FON        3000            //Inverter NTC voltage to turn fan ON
#define S_VL_TMP_FON_H      (S_VL_TMP_FON >> S_SZ_BYTE)
#define S_VL_TMP_FON_L      (S_VL_TMP_FON & S_MK_U16_L)
#define S_VL_TMP_FOF        3400            //Inverter NTC voltage to turn fan OFF
#define S_VL_TMP_FOF_H      (S_VL_TMP_FOF >> S_SZ_BYTE)
#define S_VL_TMP_FOF_L      (S_VL_TMP_FOF & S_MK_U16_L)
#define S_IN_CLIM           525             //Loaded into DAC for current limiting
#define S_IN_CLIM_H         (S_IN_CLIM >> S_SZ_BYTE)
#define S_IN_CLIM_L         (S_IN_CLIM & S_MK_U16_L)

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                         Hack Notes for Constants
 * 
 * S_LM_PWR_BOOST:  trailing zero on multiplier / divide is dropped to work around
 * compiler treating as signed 16-bit integer.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Analog channel mapping:  channel=port=pin.  The values assigned to these
 * constants should match those used by ADC.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define ANLCAN0     0x00                //AN0=RA0=2
#define ANLCAN1     0x04                //AN1=RA1=3
#define ANLCAN2     0x08                //AN2=RA2=4
#define ANLCAN3     0x0C                //AN3=RA3=5
#define ANLCAN4     0x10                //AN4=RA5=7
#define ANLCAN8     0x20                //AN8=RB2=23
#define ANLCAN9     0x24                //AN9=RB3=24
#define ANLCAN10    0x28                //AN10=RB1=22
#define ANLCAN11    0x2C                //AN11=RB4=25
#define ANLCAN12    0x30                //AN12=RB0=21
#define ANLCAN13    0x34                //AN13=RB5=26
#define ANLCAN14    0x38                //AN14=RC2=13
#define ANLCAN15    0x3C                //AN15=RC3=14
#define ANLCAN16    0x40                //AN16=RC4=15
#define ANLCAN17    0x44                //AN17=RC5=16
#define ANLCAN18    0x48                //AN18=RC6=17
#define ANLCAN19    0x4C                //AN19=RC7=18
#define ANLCANS1    0x84                //Switched AN1=RA1=3
#define ANLCANS10   0xA8                //Switched AN10=RB1=22
#define ANLCANS18   0xC8                //Switched AN18=RC6=17


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * PPS input pin definitions.  Map for MPU and use to assign peripheral input
 * pins.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define PPSRA0      0x00                //RA0
#define PPSRA1      0x01                //RA1
#define PPSRA2      0x02                //RA2
#define PPSRA3      0x03                //RA3
#define PPSRA4      0x04                //RA4
#define PPSRA5      0x05                //RA5
#define PPSRA6      0x06                //RA6
#define PPSRA7      0x07                //RA7
#define PPSRB0      0x08                //RB0
#define PPSRB1      0x09                //RB1
#define PPSRB2      0x0A                //RB2
#define PPSRB3      0x0B                //RB3
#define PPSRB4      0x0C                //RB4
#define PPSRB5      0x0D                //RB5
#define PPSRB6      0x0E                //RB6
#define PPSRB7      0x0F                //RB7
#define PPSRC0      0x10                //RC0
#define PPSRC1      0x11                //RC1
#define PPSRC2      0x12                //RC2
#define PPSRC3      0x13                //RC3
#define PPSRC4      0x14                //RC4
#define PPSRC5      0x15                //RC5
#define PPSRC6      0x16                //RC6
#define PPSRC7      0x17                //RC7


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                           PPS & direct pin mapping
 * 
 * For PPS output pin assignments both aliases for pin registers and constants 
 * mapping what peripheral connects to the pin.
 *      Pin register alias convention:  PPS_P_XXX_YY where
 *          XXX=peripheral name
 *          YY=channel or other detail; can be omitted
 *      Peripheral mapping convention:  PPS_M_XXX_YY same as alias
 * 
 * Input pins only require mapping constant.
 * 
 * For pins accessed directly or assigned in a peripheral's register (like ADC)
 * use the following convention:  PIN_X_YYYYYY where
 *          X=Input or Output
 *          YYYYYY=user assigned mnemonic describing function
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Inverter bridge gate drive pin mapping
#define PPS_P_SIDE_A_H  RC7PPS          //COG1 A maps to bridge side A high, pin 18
#define PPS_P_SIDE_A_L  RC6PPS          //COG1 D maps to bridge side A low, pin 17
#define PPS_P_SIDE_B_H  RC4PPS          //COG1 B maps to bridge side B high, pin 15
#define PPS_P_SIDE_B_L  RC5PPS          //COG1 C maps to bridge side B low, pin 16
#define PPS_M_COG1_A    0x05            //Output COG1 A signal on selected pin
#define PPS_M_COG1_B    0x06            //Output COG1 B signal on selected pin
#define PPS_M_COG1_C    0x07            //Output COG1 C signal on selected pin
#define PPS_M_COG1_D    0x08            //Output COG1 D signal on selected pin
#define PPS_M_TMR_5G    PPSRC4          //Timer5 gate pin=RC4, for getCOGtime()

//Positive (+) pole solid state relay control
#define PPS_P_PWM3_SSR  RA0PPS          //PWM3 maps to SSR gate drive on pin 2
#define PPS_M_PWM3_SSR  0x19            //Output PWM3 signal on selected pin

#define PPS_M_LAT       0x00            //Remove peripheral from pin & output latch value

//Analog input pin mapping
#define PIN_I_IVDCPV    ANLCAN13        //ADC sense - INV_+_SENSE
#define PIN_I_IVVCCV    ANLCAN12        //ADC sense - inverter VCC
#define PIN_I_IVCSSH    ANLCAN10        //ADC sense - inverter current sense S&H
#define PIN_I_INVTMP    ANLCAN8         //ADC sense - inverter temp

//Direct control input / output pins
#define PIN_O_SIDE_A_L  RC6             //Direct control of side A low switch
#define PIN_O_SIDE_B_L  RC5             //Direct control of side B low switch
#define PIN_O_SINE      RA2             //Command on for sine
#define PIN_I_SINE      RA3             //Acknowledge for sine
#define PIN_O_FAN       RA1             //Cooling fans on / off control

//LED output pins for fault, run, and power levels
#define PIN_O_FAULT     RA7             //Fault
#define PIN_O_RUN       RA5             //Run
#define PIN_O_BOOST     RC3             //Boost power
#define PIN_O_L100      RC2             //100% power
#define PIN_O_L75       RC1             //75% power
#define PIN_O_L50       RC0             //50% power
#define PIN_O_L25       RA6             //25% power


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Port/pin configuration:  I/O, analog/digital, slew rate, etc.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Port I/O:  1=input, 0=output
#define IO_PORTA    0b00001000
#define IO_PORTB    0b10101101
#define IO_PORTC    0b00000000
//Port Analog/Digital:  1=analog, 0=digital
#define AD_PORTA    0b00000000
#define AD_PORTB    0b00101111
#define AD_PORTC    0b00000000
//Port weak pull-up:  1=enabled, 0=disabled
#define WP_PORTA    0b00000000
#define WP_PORTB    0b00000000
#define WP_PORTC    0b00000000
//Open drain:  1=open drain, 0=push-pull
#define OD_PORTA    0b00000000
#define OD_PORTB    0b00000000
#define OD_PORTC    0b00000000
//Slew rate:  1=limited, 0=maximum
#define SR_PORTA    0b00000000
#define SR_PORTB    0b00000000
#define SR_PORTC    0b00000000
//Input level:  1=ST, 0=TLL
#define IL_PORTA    0b00000000
#define IL_PORTB    0b00000000
#define IL_PORTC    0b00000000
//Initial latch states:  all outputs off.
#define S_IN_PORTA  0b00000000
#define S_IN_PORTB  0b00000000
#define S_IN_PORTC  0b00000000


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                            Macro functions
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define clearCmdResult()    cmdGood = TRUE
#define setCmdOK()          cmdGood = TRUE
#define setCmdErr()         cmdGood = FALSE
#define relayStatus()       PWM3DCH
#define invStatus()         ((COG1ASD0bits.ASE) ? FALSE : TRUE)
#define sineStatus()        PIN_O_SINE
#define sineAck()           PIN_I_SINE

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * These macros only function if prnResult has been set by console service in 
 * the main run-time loop.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define printCmdResult()    if (prnResult) if (cmdGood) printf(msgOk); else printf(msgErr)
#define printTogResult(ts)  if (prnResult) if (cmdGood) printf(msgTog, ((ts) ? msgOn : msgOff)); else printf(msgErr)
#define printCmdSys()       if (prnResult) printf("%s%s\n", lblSys, StsList[Config.sts].desc)

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Used by several console functions to remove whitespace between command and
 * parameter.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define rmvSpaces(cs)       for (cs += CMDSIZSTR; (*cs == CMDWSPACE); ++cs)


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *              ADC configuration & acquisition constants
 * 
 * IMPORTANT:  many of these constants are ratios or quantities dependent on 
 * the ADC's Vref value.  This value should be maintained in these comments
 * and all constants reviewed if it is changed.
 * 
 * Vref=4.096
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//AD acquisition configuration & control constants
#define CFG_ADC_CON_1       0xD3            //Result right justified, Fosc/16, Vss, FVR
#define CFG_ADC_CON_2       0x00            //No auto conversion
#define S_LM_ADC_VAL        1023            //Maximum value 10-bit ADC can return
#define S_QY_ADC_1V         250             //ADC value for 1V
#define S_RT_ADC_MV         4               //Millivolts per ADC bit
#define S_TM_ADC_ACQ        20              //# uS required for sample & hold cap to acquire
#define S_SZ_ADC_INT        sizeof(unsigned int)    //# bytes in ADC integer
#define POS_ADC_HB          1               //ADC result high byte
#define POS_ADC_LB          0               //ADC result low byte

//Internal channels (peripherals)
#define ADC_CH_FVR          0xFC            //FVR
#define ADC_CH_DAC_1        0xF8            //DAC1
#define ADC_CH_TMP          0xF4            //Temperature indicator
#define ADC_CH_DAC_2        0xF0            //DAC2
#define ADC_CH_DAC_3        0xEC            //DAC3
#define ADC_CH_DAC_4        0xE8            //DAC4
#define ADC_CH_DAC_5        0xE4            //DAC5
#define ADC_CH_DAC_7        0xDC            //DAC7


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Flash program memory definitions.  For non-volatile storage of configurable
 * parameters.
 * 
 * Important:  flash architecture is PIC family dependent and the memory map /
 * constants must be reviewed when changing MCUs.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define FLASHROWSZ          32              //Size of flash row: PIC specific
#define FLASH_SZ_CSUM   (FLASHROWSZ - 2)    //Store checksum in last word of row

//Flash memory unlock sequence macro
#define unlockSeq() PMCON2 = 0x55; PMCON2 = 0xAA; WR = TRUE; NOP(); NOP()

//Row address for data:  this value must match default assignment below
#define FLASH_AR_AUTO       0               //Autostart setting
#define FLASH_AR_STS        1               //Last status code
#define FLASH_AR_OLDC       2               //PWM open loop DC register value
#define FLASH_AR_OLDC_P     3               //Open loop DC %
#define FLASH_AR_CLIM_L     4               //Current limit (clim) low byte
#define FLASH_AR_CLIM_H     5               //Current limit (clim) high byte
#define FLASH_AR_BTIM       6               //COG rising edge blanking time of falling events
#define FLASH_AR_FAN_O_L    7               //NTC voltage to turn fan on low byte
#define FLASH_AR_FAN_O_H    8               //NTC voltage to turn fan on high byte
#define FLASH_AR_FAN_F_L    9               //NTC voltage to turn fan off low byte
#define FLASH_AR_FAN_F_H    10              //NTC voltage to turn fan off hight byte
#define FLASH_AR_OVH_L      11              //NTC voltage for shutdown low byte
#define FLASH_AR_OVH_H      12              //NTC voltage for shutdown high byte
#define FLASH_AR_ICS        13              //Inverter amperes reduction factor

//Checksum stored in little endian format in last word of row
#define FLASH_AR_CSUM_L (FLASHROWSZ - 2)
#define FLASH_AR_CSUM_H (FLASHROWSZ - 1)

//Default data to initialize row with.  Positions 30 & 31 reserved for checksum.
#define S_DF_FLASH_0        S_BL_AUTOD
#define S_DF_FLASH_1        STS_UND
#define S_DF_FLASH_2        S_DC_RUN
#define S_DF_FLASH_3        S_DC_RUN_P
#define S_DF_FLASH_4        S_IN_CLIM_L
#define S_DF_FLASH_5        S_IN_CLIM_H
#define S_DF_FLASH_6        CFG_PWM_11_DUTYC
#define S_DF_FLASH_7        S_VL_TMP_FON_L
#define S_DF_FLASH_8        S_VL_TMP_FON_H
#define S_DF_FLASH_9        S_VL_TMP_FOF_L
#define S_DF_FLASH_10       S_VL_TMP_FOF_H
#define S_DF_FLASH_11       S_VL_TMP_OVH_L
#define S_DF_FLASH_12       S_VL_TMP_OVH_H
#define S_DF_FLASH_13       S_RT_ICS_AMPS
#define S_DF_FLASH_14       0
#define S_DF_FLASH_15       0
#define S_DF_FLASH_16       0
#define S_DF_FLASH_17       0
#define S_DF_FLASH_18       0
#define S_DF_FLASH_19       0
#define S_DF_FLASH_20       0
#define S_DF_FLASH_21       0
#define S_DF_FLASH_22       0
#define S_DF_FLASH_23       0
#define S_DF_FLASH_24       0
#define S_DF_FLASH_25       0
#define S_DF_FLASH_26       0
#define S_DF_FLASH_27       0
#define S_DF_FLASH_28       0
#define S_DF_FLASH_29       0


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer2 configuration:  time base for PWM3.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_TMR_2_CLK   0x01            //Fosc/4
#define CFG_TMR_2_CON   0x80            //Enabled, 1:1 prescale, 1:1 post scale
#define CFG_TMR_2_HLT   0x00            //No sync, free running
#define CFG_TMR_2_RST   0x00            //No external reset sources
#define CFG_TMR_2_PR    39              //100kHz

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer5 configuration and control constants for function getCOGtime.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_TMR_5_C     0x41            //Fosc, 1:1 prescale, sync, on
#define CFG_TMR_5_G     0xD0            //Gate enable, 1 polarity, single pulse, gate pin

#define FGCT_MAXCNT     160             //Maximum count for 10uS
#define FGCT_1USCNT     16              //Count for 1uS
#define FGCT_WAITTM     30              //# uS to wait before checking go/done bit

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Timer8 configuration:  provides interrupt timing for background caretaker that
 * performs system monitor & control.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_TMR_8_CLK   0x05            //MFINTOSC
#define CFG_TMR_8_CON   0x19            //Off, 1:2 prescale, 1:10 post scale
#define CFG_TMR_8_HLT   0xA0            //Prescale syncs Fosc/4, on bit syncs Fosc/4, free running
#define CFG_TMR_8_RST   0x00            //No external reset sources
#define CFG_TMR_8_PR    250             //Period value to obtain 10mS period


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * FVR, DAC, and function Initialize configuration constants
 * DAC1:  bridge over-current sense
 * DAC4:  Vdd sense
 * DAC5:  OPEN
 * DAC7:  OPEN
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_FVR_C       0xCF            //Enabled, no temp sense, CDAFVRx4, ADFVRx4

//DAC1:  voltage is set from NVM
#define CFG_DAC_1_C     0x88            //Enabled, right justified, FVR buffer 2, Vss

//DAC4:  Vdd/2
#define CFG_DAC_4_C     0x80            //Enabled, Vdd, Vss
#define CFG_DAC_4_V     0x10            //Output byte

#define FI_FVR_DLY      30              //Time delay in uS for FVR buffer to stabilize


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Current sense sample & hold OPA2 configuration
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_OPA_2_C     0x11            //Disabled, unity gain, override tri-states output
#define CFG_OPA_2_O     0x15            //CLC2 is override source
#define CFG_OPA_2_N     0x00            //Not used with unity gain
#define CFG_OPA_2_P     0x00            //Connect (+) to OPA2IN0+


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure comparator 2 for inverter bridge over-current sense.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CM_2_C0     0xAA            //Enabled, normal pwr, ZLF on, HYS on
#define CFG_CM_2_C1     0x01            //Interrupt on falling edge
#define CFG_CM_2_N      0x02            //Connect (-) to C2IN2- pin (24)
#define CFG_CM_2_P      0x05            //Connect (+) to DAC1 output


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure 10-bit PWM3:  gate drive for SSR (relay)
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PWM_3_CON       0x80            //Enabled, active high output
#define CFG_PWM_3_DCL       0x00            //2 LSB not used
#define CFG_PWM_3_DCH_OFF   0x00            //Relay off
#define CFG_PWM_3_DCH_ON    0x05            //Relay on, ~ 1.25uS DC


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure 16-bit PWM5:  master and sets max DC
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PWM_5_CON      0x80                 //Enabled, active high output, standard mode
#define CFG_PWM_5_INTE     0x00                 //No interrupts
#define CFG_PWM_5_CLK      0x01                 //No prescale, use HFINTOSC as clock source
#define CFG_PWM_5_LDCON    0x80                 //Load triggering is immediate
#define CFG_PWM_5_OFCON    0x00                 //Offset trigger source is independent run mode
#define CFG_PWM_5_PHASE    0x00                 //Phase count
#define CFG_PWM_5_PERIOD   S_TM_INV_F           //Period (frequency)
#define CFG_PWM_5_OFFSET   0x0000               //Offset count
#define CFG_PWM_5_TIMER    0x0000               //Starting value for timer


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure PWM11:  slave to PWM5 to time current sense falling event blanking
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_PWM_11_CON      0x80                //Enabled, active high output, standard mode
#define CFG_PWM_11_INTE     0x00                //No interrupts
#define CFG_PWM_11_CLK      0x01                //No prescale, use HFINTOSC as clock source
#define CFG_PWM_11_LDCON    0x80                //Load buffers at end of current period
#define CFG_PWM_11_OFCON    0x20                //Slave run mode with offset triggered sync start from PWM5
#define CFG_PWM_11_PHASE    0x00                //Phase
#define CFG_PWM_11_DUTYC    0x000D              //Duty cycle (blanking time - 812nS)
#define CFG_PWM_11_PERIOD   S_TM_INV_F          //Period
#define CFG_PWM_11_OFFSET   0x0000              //Offset count
#define CFG_PWM_11_TIMER    0x0000              //Starting value for timer


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure CLC2 as OR gate to control OPA2 tri-state for current sense S&H
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CLC_2_CON       0x00                //Disabled, no interrupts, AND-OR
#define CFG_CLC_2_POL       0x00                //No inverted polarities
#define CFG_CLC_2_SEL_0     0x10                //COG1A selected for d1T/g1
#define CFG_CLC_2_SEL_1     0x10                //COG1A selected for d2T/g2
#define CFG_CLC_2_SEL_2     0x11                //COG1B selected for d3T/g3
#define CFG_CLC_2_SEL_3     0x11                //COG1B selected for d4T/g4
#define CFG_CLC_2_GLS_0     0x02                //d1T is gated into g1
#define CFG_CLC_2_GLS_1     0x08                //d2T is gated into g2
#define CFG_CLC_2_GLS_2     0x20                //d3T is gated into g3
#define CFG_CLC_2_GLS_3     0x80                //d4T is gated into g4

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure CLC3 as OR gate to control current sense blanking
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_CLC_3_CON       0x00                //Disabled, no interrupts, AND-OR
#define CFG_CLC_3_POL       0x00                //No inverted polarities
#define CFG_CLC_3_SEL_0     0x22                //PWM11 selected for d1T/g1
#define CFG_CLC_3_SEL_1     0x22                //PWM11 selected for d2T/g2
#define CFG_CLC_3_SEL_2     0x09                //CM2 selected for d3T/g3
#define CFG_CLC_3_SEL_3     0x09                //CM2 selected for d4T/g4
#define CFG_CLC_3_GLS_0     0x02                //d1T is gated into g1
#define CFG_CLC_3_GLS_1     0x08                //d2T is gated into g2
#define CFG_CLC_3_GLS_2     0x20                //d3T is gated into g3
#define CFG_CLC_3_GLS_3     0x80                //d4T is gated into g4


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Configure Complementary Output Generator 1 (COG1).  Start with peripheral
 * config.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_COG_1_CON0          0xD5            //Enable, reg load next event, HFINTOSC clock, push/pull
#define CFG_COG_1_CON1          0x00            //Clock used for dead band, all outputs active high

//Rising events
#define CFG_COG_1_REI0_NONE     0x00            //No rising event inputs for RIS0
#define CFG_COG_1_REI1_PWM5     0x02            //PWM5 rising event input for RIS1
#define CFG_COG_1_REM0_NONE     0x00            //No rising event for RSIM0
#define CFG_COG_1_REM1_PWM5     0x02            //PWM5 low->high for RSIM1

//Falling events
#define CFG_COG_1_FEI0_NONE     0x00            //No falling events
#define CFG_COG_1_FEI1_OLOOP    0x42            //Open loop:  CLC3 & PWM5 FE input
#define CFG_COG_1_FEM0_NONE     0x00            //N/A
#define CFG_COG_1_FEM1_OLOOP    0x42            //PWM5, CLC3: high->low
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Explanation of CLC3 FE set to high->low.  The blanking time scheme blocks 
 * over-current events at the start of the cycle for a short
 * period of time, expected to be << 1uS.  If a true over-current event occurs
 * during the blanking time it's critical for the DC to end as soon as the
 * blanking time expires.
 * 
 * The blanking time configuration will set the over-current FE signal high
 * at the start of the cycle regardless of the comparator state.  At the end of
 * the blanking time and if the comparator state is set low (indicating an over-
 * current event), the FE will go low thus terminating the DC.
 * 
 * A low state is not used because of the noise in the current-sense circuit and
 * propogation delays in the blanking time configuration.  It's possible with 
 * high DCs (>90%) that the ringing from the previous cycle triggers the
 * comparator.  Due to the propogation delays at the start of the next cycle if
 * the CLC output can still be low and prevent a legitimate cycle from starting.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

//Auto shutdown
#define CFG_COG_1_ASD0          0xA8            //COG is shutdown, all outputs set low
#define CFG_COG_1_ASD1          0x00            //No auto-shutdown sources

//Delays & blanking
#define CFG_COG_1_RE_DEADB      0x00            //No rising dead band
#define CFG_COG_1_FE_DEADB      0x00            //No falling dead band
#define CFG_COG_1_RE_BLANK      0x00            //No rising edge blanking
#define CFG_COG_1_FE_BLANK      0x00            //No falling edge blanking
#define CFG_COG_1_RE_PHASE      0x00            //No rising event phase delay
#define CFG_COG_1_FE_PHASE      0x00            //No falling event phase delay


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                          Print & Format Constants
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define FMT_HELP_INTRO          "\n%s\nHW model: %s, revision: %s\nFW version \
%u.%u, build %u\n%s\n\nCommands:\n"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                          Object definitions
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Object to store a measured voltage with 3-digit mV resolution
typedef struct {
    unsigned int volt;                  //Voltage
    unsigned int mv;                    //Millivolts (3 decimals)
    unsigned int vmv;                   //Composite volts & mv represented as integer
    unsigned int adc;                   //Unmodified value read from ADC
} volts;

//Configuration record to control operation of inverter
//To calculate DC% from measured DC:  Config.odcp = ((Config.oldc * 100) / (CFG_PWM_5_PERIOD + 1));
typedef struct {
    unsigned char sts;                  //System status code; see status codes
    unsigned char cogt;                 //Calculated COG on-time derived from function getCOGtime
    unsigned char icsr;                 //Integer representing % to reduce ics to derive input amps
    unsigned int  cogc;                 //Timer count of COG duty cycle
    unsigned int  odcp;                 //Open loop duty cycle % entered by user
    unsigned int  oldc;                 //Calculated 16-bit open loop duty cycle
    unsigned int  btim;                 //PWM11 duty cycle value (CS blanking time)
    unsigned int  amps;                 //Calculated inverter primary current
    unsigned int  pwr;                  //Calculated inverter primary power
    unsigned int  pwr_e;                //Calculated inverter equivalent primary power
    unsigned int  clim;                 //DAC voltage reference for current limit
    unsigned int  tfon;                 //NTC voltage representing temp to turn fan on
    unsigned int  tfof;                 //NTC voltage representing temp to turn fan off
    unsigned int  tovh;                 //NTC voltage representing overheat & shutdown
    volts inv;                          //Inverter DC+ bus voltage
    volts vcc;                          //Inverter auxillary (Vcc) voltage
    volts vdd;                          //Inverter logic (Vdd) voltage)
    volts ics;                          //Inverter current sense S&H voltage
    volts tmp;                          //Inverter NTC voltage (temp)
} invcf;


typedef struct {
    const char *str;
    const char *desc;
} cmd;

const cmd CmdList[] = {
    {"CE","Clear nvm Error"},
    {"DS","Display Status"},
    {"GA","Amps"},
    {"GD","DC"},
    {"GT","Temp"},
    {"GV","Volts"},
    {"GW","Watts"},
    {"RD","Defaults"},
    {"RT","Reset"},
    {"SA","Autostart <D/I/S>"},
    {"SB","Blanking <#>"},
    {"SC","Current limit <#>"},
    {"SD","Duty cycle <%>"},
    {"SE","Set Echo #"},
    {"SR","Ratio amps %"},
    {"TF","Off @ #"},
    {"TO","On @ #"},
    {"TS","Shutdown @ #"},
    {"XF","Fan"},
    {"XI","Inverter"},
    {"XM","Monitor"},
    {"XR","Relay"},
    {"XS","Sine"},
    {"?","Help"}
};
#define NUMOFCMDS   sizeof(CmdList) / (sizeof(const char *) * 2)
#define CMDCLRERR   0
#define CMDDSPSTS   1
#define CMDGETAMP   2
#define CMDGETDYC   3
#define CMDGETEMP   4
#define CMDGETVOL   5
#define CMDGETWAT   6
#define CMDRSTDFL   7
#define CMDRSTPIC   8
#define CMDSETAUT   9
#define CMDSETBLK   10
#define CMDSETCUR   11
#define CMDSETDCY   12
#define CMDSETECH   13
#define CMDSETRAT   14
#define CMDTMPFOF   15
#define CMDTMPFON   16
#define CMDTMPOVH   17
#define CMDTOGFAN   18
#define CMDTOGINV   19
#define CMDTOGMON   20
#define CMDTOGRLY   21
#define CMDTOGSIN   22
#define CMDDSPHLP   23

#define CMDSTRING   0                   //Array position for command string
#define CMDDESCRP   1                   //Array position for command description
#define CMDSIZSTR   2                   //Size of command string
#define CMDLOWCAS   32                  //Add this to upper case to make lower
#define CMDWSPACE   ' '                 //White space
#define CMDCLOSLP   'c'                 //Closed loop parameter
#define CMDOPENLP   'o'                 //Open loop parameter
#define CMDENABLE   'e'                 //Enable
#define CMDDSABLE   'd'                 //Disable
#define CMDINVERT   'i'                 //Auto-start inverter only
#define CMDINSINE   's'                 //Auto-start inverter & sine
#define CMDLOOPAN   'a'                 //Analog loop
#define CMDLOOPDG   'd'                 //Digital loop
#define CMDECHOOF   '0'                 //Console character echo off


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                           STATUS CODES & LED MAP
 * 
 * The constants STS_XXXX represent the status codes used throughout the 
 * control program and stored in object Config.sts as well as recorded in NVM.
 * 
 * The object StsList provides the LED bit-map for the front panel display of
 * diagnostic codes and brief text for status display via the service port.
 * 
 * The byte StsList[].ledmap is a bit-map code to display unit status using
 * the front panel LEDs and works as follows.
 *      + The load LEDs are expected to be arranged left to right (facing them):
 *          25  50  75  100/Boost
 *      + The Run/Fault LED can be placed to the left or right of the load LEDs.
 *      + Bit mapping and respective LED colors:
 *          - 0=Fault (red)
 *          - 1=Run (green)
 *          - 2=Boost (yellow)
 *          - 3=100 (green)
 *          - 4=75 (green)
 *          - 5=50 (green)
 *          - 6=25 (green)
 *      + The Run/Fault and 100/Boost are single bi-color LEDs (anti-parallel
 *        diodes) so one pin must be on and the other off.
 *      + The LED pattern forms a 4-bit BCD (binary counting decimal) code
 *        equal to the status code.
 *      + For codes < 16 the 100% (bit 3) is used to represent bit 0 of the 
 *        code.
 *      + For codes 16-23 the boost pin (bit 2) is set on and bits 4-6 represent
 *        the value.  Example:  a status code of 16 would set the boost LED on
 *        and the 25-75% LEDs off.  A value of 18 would set the boost & 50%
 *        LEDs on.
 *      + The 'RUN' status code is unique in program logic and differentiates
 *        between displaying a status code and power so, even though the map
 *        is populated it doesn't display.
 *      + The design workbook contains a LED mapping code calculator on the
 *        'ledmap' tab that generates the mapping value to place in ledmap for
 *        each respective status code.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
typedef struct {
    const unsigned char ledmap;
    const char *desc;
} sts;

#define STS_STOP        0               //Stopped
#define STS_RUN         1               //Running
#define STS_BOOT        2               //Booting
#define STS_BUS_UVLO    3               //Low inverter input voltage
#define STS_OVC_OCLO    4               //Inverter current draw too high
#define STS_CHL_BSAF    5               //Inverter A-side bootstrap charge failed
#define STS_CHL_BSBF    6               //Inverter B-side bootstrap charge failed
#define STS_CPU         7               //MCU fault (stack, watchdog, etc.)
#define STS_COG_TIME    8               //Low COG time during inverter start
#define STS_SIN_CMD     9               //Sine failure (no acknowledgement)
#define STS_NVM_INV     10              //Invalid NVM
#define STS_HIGH_TEMP   11              //Inverter high temp
#define STS_OVERLOAD    12              //Power overload
#define STS_BUS_OVLO    13              //High inverter input voltage
#define STS_UND         14              //Undefined: must be last in list

#define STS_CODE_ERR    STS_BUS_UVLO    //Codes < value ARE NOT errors
#define STS_CODE_OK     (STS_CODE_ERR - 1)  //Codes > value ARE errors

const sts StsList[] = {
    {0x01,"stop"},
    {0x09,"run"},
    {0x11,"boot"},
    {0x19,"UVLO"},
    {0x21,"OCLO"},
    {0x29,"BS A"},
    {0x31,"BS B"},
    {0x39,"CPU"},
    {0x41,"COG"},
    {0x49,"sine"},
    {0x51,"NVM"},
    {0x59,"temp"},
    {0x61,"overload"},
    {0x69,"OVLO"},
    {0x75,"UNDF"}
};


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Message, format, and label strings
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
const char *msgName = "True Sine Inverter",
        *msgMod = "TS350r1",
        *msgHwr = "190314",
        *msgCW = "Copyright (c) 2019 Brian Cornell",
        *msgErr = "ERR\n",
        *msgHuh = "\n?\n",
        *msgOk = "Ok\n",
        *msgNewl = "\n\n",
        *msgOn = "on",
        *msgOff = "off",
        *msgInv = "inverter",
        *msgSine = "sine",
        *lblInv = "Inv=",
        *lblTemp = "Tmp=",
        *lblDc = "DC=",
        *fmtMonvd = "%s%u.%3.3u\t",
        *fmtMonlt = "%s%u.%3.3u\n",
        *fmtGetv = "%u\n",
        *fmtGetvd = "%u.%3.3u\n",
        *lblSys = "Sys: ",
        *lblAmp = "A=",
        *lblPwr = "W=",
        *msgTog = "%s\n";


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                     Configure system clock & Watchdog Timer
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_OSC_C       0x78            //16MHz, FOSC config word determines clock
#define CFG_WDT_CON     0x10            //Watchdog interval=256mS, off.


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Function declarations.  See source file for explanation of each function.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void interrupt InterruptExec(void);
void Initialize(void);
void main(void);
void DisplayStatus(void);
void setDutyCycle(char *);
void setCurrentLimit(char *);
void setVref(char *);
bit InverterON(void);
void InverterOFF(void);
void ReadVoltage(unsigned char, unsigned char, volts *);
void CalcAvg(unsigned int, unsigned char, unsigned char, volts *);
void getCOGtime(void);
void chgDutyCycle(unsigned int, unsigned int);
unsigned char chargeBScaps(void);
bit setSine(unsigned char);
void setRelay(unsigned char);
void eraseFlashRow(uint16_t addr);
void writeFlashRow(uint16_t, unsigned char *);
unsigned int nvmChksum(unsigned char *);
void setAutostart(char *);
void setBlanking(char *);
void setTemps(char *, unsigned char);
void setAmpsReduction(char *);


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Preprocessor checks.  Maintain at end of main include file.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Check for maximum current sense sample size exceeding unsigned int.
#define _ADC_OVC_VAL    ((S_IN_CLIM_H << 8) | S_IN_CLIM_L)
#if((_ADC_OVC_VAL * S_LM_SAMPLES) > UINT_MAX)
#error "Max current sense value _ADC_OVC_VAL * S_LM_SAMPLES exceeds unsigned int."
#endif



//End of file